# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import aiohttp
import asyncio
import unittest

from instana.singletons import async_tracer, agent

import tests.apps.flask_app
import tests.apps.aiohttp_app
from ..helpers import testenv


class TestAiohttp(unittest.TestCase):

    async def fetch(self, session, url, headers=None):
        try:
            async with session.get(url, headers=headers) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = async_tracer.recorder
        self.recorder.clear_spans()

        # New event loop for every test
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

    def tearDown(self):
        pass

    def test_client_get(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, wsgi_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(wsgi_span.ec)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + "/",
                         aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        assert "X-INSTANA-T" in response.headers
        self.assertEqual(response.headers["X-INSTANA-T"], traceId)
        assert "X-INSTANA-S" in response.headers
        self.assertEqual(response.headers["X-INSTANA-S"], wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        self.assertEqual(response.headers["X-INSTANA-L"], '1')
        assert "Server-Timing" in response.headers
        self.assertEqual(
            response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_client_get_301(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/301")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(4, len(spans))

        wsgi_span1 = spans[0]
        wsgi_span2 = spans[1]
        aiohttp_span = spans[2]
        test_span = spans[3]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, wsgi_span1.t)
        self.assertEqual(traceId, wsgi_span2.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(wsgi_span1.p, aiohttp_span.s)
        self.assertEqual(wsgi_span2.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(wsgi_span1.ec)
        self.assertIsNone(wsgi_span2.ec)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + "/301",
                         aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        assert "X-INSTANA-T" in response.headers
        self.assertEqual(response.headers["X-INSTANA-T"], traceId)
        assert "X-INSTANA-S" in response.headers
        self.assertEqual(response.headers["X-INSTANA-S"], wsgi_span2.s)
        assert "X-INSTANA-L" in response.headers
        self.assertEqual(response.headers["X-INSTANA-L"], '1')
        assert "Server-Timing" in response.headers
        self.assertEqual(
            response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_client_get_405(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/405")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, wsgi_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(wsgi_span.ec)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual(405, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + "/405",
                         aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        assert "X-INSTANA-T" in response.headers
        self.assertEqual(response.headers["X-INSTANA-T"], traceId)
        assert "X-INSTANA-S" in response.headers
        self.assertEqual(response.headers["X-INSTANA-S"], wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        self.assertEqual(response.headers["X-INSTANA-L"], '1')
        assert "Server-Timing" in response.headers
        self.assertEqual(
            response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_client_get_500(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/500")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, wsgi_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(aiohttp_span.ec, 1)
        self.assertEqual(wsgi_span.ec, 1)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual(500, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + "/500",
                         aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertEqual('INTERNAL SERVER ERROR',
                         aiohttp_span.data["http"]["error"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        assert "X-INSTANA-T" in response.headers
        self.assertEqual(response.headers["X-INSTANA-T"], traceId)
        assert "X-INSTANA-S" in response.headers
        self.assertEqual(response.headers["X-INSTANA-S"], wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        self.assertEqual(response.headers["X-INSTANA-L"], '1')
        assert "Server-Timing" in response.headers
        self.assertEqual(
            response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_client_get_504(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/504")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, wsgi_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(aiohttp_span.ec, 1)
        self.assertEqual(wsgi_span.ec, 1)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual(504, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + "/504",
                         aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertEqual('GATEWAY TIMEOUT', aiohttp_span.data["http"]["error"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        assert "X-INSTANA-T" in response.headers
        self.assertEqual(response.headers["X-INSTANA-T"], traceId)
        assert "X-INSTANA-S" in response.headers
        self.assertEqual(response.headers["X-INSTANA-S"], wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        self.assertEqual(response.headers["X-INSTANA-L"], '1')
        assert "Server-Timing" in response.headers
        self.assertEqual(
            response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_client_get_with_params_to_scrub(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/?secret=yeah")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, wsgi_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(wsgi_span.ec)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + "/",
                         aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertEqual("secret=<redacted>",
                         aiohttp_span.data["http"]["params"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        assert "X-INSTANA-T" in response.headers
        self.assertEqual(response.headers["X-INSTANA-T"], traceId)
        assert "X-INSTANA-S" in response.headers
        self.assertEqual(response.headers["X-INSTANA-S"], wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        self.assertEqual(response.headers["X-INSTANA-L"], '1')
        assert "Server-Timing" in response.headers
        self.assertEqual(
            response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_client_response_header_capture(self):
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ['X-Capture-This']

        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/response_headers")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, wsgi_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(wsgi_span.ec)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + "/response_headers", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        assert "X-Capture-This" in aiohttp_span.data["http"]["header"]
        self.assertEqual("Ok", aiohttp_span.data["http"]["header"]["X-Capture-This"])

        assert "X-INSTANA-T" in response.headers
        self.assertEqual(response.headers["X-INSTANA-T"], traceId)
        assert "X-INSTANA-S" in response.headers
        self.assertEqual(response.headers["X-INSTANA-S"], wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        self.assertEqual(response.headers["X-INSTANA-L"], '1')
        assert "Server-Timing" in response.headers
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

        agent.options.extra_http_headers = original_extra_http_headers

    def test_client_error(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, 'http://doesnotexist:10/')

        response = None
        try:
            response = self.loop.run_until_complete(test())
        except:
            pass

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        aiohttp_span = spans[0]
        test_span = spans[1]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(aiohttp_span.ec, 1)

        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertIsNone(aiohttp_span.data["http"]["status"])
        self.assertEqual("http://doesnotexist:10/",
                         aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.data["http"]["error"])
        assert(len(aiohttp_span.data["http"]["error"]))
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertIsNone(response)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

"""
This file contains a config object that will hold configuration options for the package.
Defaults are set and can be overridden after package load.
"""
from __future__ import absolute_import
from .util import DictionaryOfStan

# La Protagonista
config = DictionaryOfStan()


# This option determines if tasks created via asyncio (with ensure_future or create_task) will
# automatically carry existing context into the created task.
config['asyncio_task_context_propagation']['enabled'] = False







# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
The in-process Instana agent (for testing & the test suite) that manages
monitoring state and reporting that data.
"""
import os
from ..log import logger
from .host import HostAgent


class TestAgent(HostAgent):
    """
    Special Agent for the test suite.  This agent is based on the StandardAgent.  Overrides here are only for test
    purposes and mocking.
    """
    def get_from_structure(self):
        """
        Retrieves the From data that is reported alongside monitoring data.
        @return: dict()
        """
        return {'e': os.getpid(), 'h': 'fake'}

    def can_send(self):
        """
        Are we in a state where we can send data?
        @return: Boolean
        """
        return True

    def report_traces(self, spans):
        logger.warning("Tried to report_traces with a TestAgent!")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import json
import sys
import time

from collections import defaultdict
import pkg_resources

try:
    from urllib import parse
except ImportError:
    import urlparse as parse
    import urllib

from ..log import logger

if sys.version_info.major == 2:
    string_types = basestring
else:
    string_types = str

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

def nested_dictionary():
    return defaultdict(DictionaryOfStan)

# Simple implementation of a nested dictionary.
DictionaryOfStan = nested_dictionary


def to_json(obj):
    """
    Convert obj to json.  Used mostly to convert the classes in json_span.py until we switch to nested
    dicts (or something better)

    :param obj: the object to serialize to json
    :return:  json string
    """
    try:
        def extractor(o):
            if not hasattr(o, '__dict__'):
                logger.debug("Couldn't serialize non dict type: %s", type(o))
                return {}
            else:
                return {k.lower(): v for k, v in o.__dict__.items() if v is not None}

        return json.dumps(obj, default=extractor, sort_keys=False, separators=(',', ':')).encode()
    except Exception:
        logger.debug("to_json non-fatal encoding issue: ", exc_info=True)

def to_pretty_json(obj):
    """
    Convert obj to pretty json.  Used mostly in logging/debugging.

    :param obj: the object to serialize to json
    :return:  json string
    """
    try:
        def extractor(o):
            if not hasattr(o, '__dict__'):
                logger.debug("Couldn't serialize non dict type: %s", type(o))
                return {}
            else:
                return {k.lower(): v for k, v in o.__dict__.items() if v is not None}

        return json.dumps(obj, default=extractor, sort_keys=True, indent=4, separators=(',', ':'))
    except Exception:
        logger.debug("to_pretty_json non-fatal encoding issue: ", exc_info=True)


def package_version():
    """
    Determine the version of this package.

    :return: String representing known version
    """
    version = ""
    try:
        version = pkg_resources.get_distribution('instana').version
    except pkg_resources.DistributionNotFound:
        version = 'unknown'

    return version


def get_default_gateway():
    """
    Attempts to read /proc/self/net/route to determine the default gateway in use.

    :return: String - the ip address of the default gateway or None if not found/possible/non-existant
    """
    try:
        hip = None
        # The first line is the header line
        # We look for the line where the Destination is 00000000 - that is the default route
        # The Gateway IP is encoded backwards in hex.
        with open("/proc/self/net/route") as routes:
            for line in routes:
                parts = line.split('\t')
                if parts[1] == '00000000':
                    hip = parts[2]

        if hip is not None and len(hip) == 8:
            # Reverse order, convert hex to int
            return "%i.%i.%i.%i" % (int(hip[6:8], 16), int(hip[4:6], 16), int(hip[2:4], 16), int(hip[0:2], 16))

    except Exception:
        logger.warning("get_default_gateway: ", exc_info=True)


def every(delay, task, name):
    """
    Executes a task every `delay` seconds

    :param delay: the delay in seconds
    :param task: the method to run.  The method should return False if you want the loop to stop.
    :return: None
    """
    next_time = time.time() + delay

    while True:
        time.sleep(max(0, next_time - time.time()))
        try:
            if task() is False:
                break
        except Exception:
            logger.debug("Problem while executing repetitive task: %s", name, exc_info=True)

        # skip tasks if we are behind schedule:
        next_time += (time.time() - next_time) // delay * delay + delay


def validate_url(url):
    """
    Validate if <url> is a valid url

    Examples:
    - "http://localhost:5000" - valid
    - "http://localhost:5000/path" - valid
    - "sandwich" - invalid

    @param url: string
    @return: Boolean
    """
    try:
        result = parse.urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception:
        pass

    return False



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import threading


def launch_background_thread(app, app_name, fun_args=(), fun_kwargs={}):
    print("Starting background %s app..." % app_name)
    app_thread = threading.Thread(target=app,
                                  name=app_name,
                                  args=fun_args,
                                  kwargs=fun_kwargs)
    app_thread.daemon = True
    app_thread.start()
    return app_thread



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import json
import unittest
import logging

from nose.tools import (assert_is_none, assert_is_not_none,
                        assert_false, assert_true, assert_list_equal)

from ..helpers import testenv
from instana.singletons import tracer

import pymongo
import bson

logger = logging.getLogger(__name__)


class TestPyMongoTracer(unittest.TestCase):
    def setUp(self):
        self.conn = pymongo.MongoClient(host=testenv['mongodb_host'], port=int(testenv['mongodb_port']),
                                        username=testenv['mongodb_user'], password=testenv['mongodb_pw'])
        self.conn.test.records.delete_many(filter={})

        self.recorder = tracer.recorder
        self.recorder.clear_spans()

    def tearDown(self):
        return None

    def test_successful_find_query(self):
        with tracer.start_active_span("test"):
            self.conn.test.records.find_one({"type": "string"})

        assert_is_none(tracer.active_span)

        spans = self.recorder.queued_spans()
        self.assertEqual(len(spans), 2)

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        assert_is_none(db_span.ec)

        self.assertEqual(db_span.n, "mongo")
        self.assertEqual(db_span.data["mongo"]["service"], "%s:%s" % (testenv['mongodb_host'], testenv['mongodb_port']))
        self.assertEqual(db_span.data["mongo"]["namespace"], "test.records")
        self.assertEqual(db_span.data["mongo"]["command"], "find")

        self.assertEqual(db_span.data["mongo"]["filter"], '{"type": "string"}')
        assert_is_none(db_span.data["mongo"]["json"])

    def test_successful_insert_query(self):
        with tracer.start_active_span("test"):
            self.conn.test.records.insert_one({"type": "string"})

        assert_is_none(tracer.active_span)

        spans = self.recorder.queued_spans()
        self.assertEqual(len(spans), 2)

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        assert_is_none(db_span.ec)

        self.assertEqual(db_span.n, "mongo")
        self.assertEqual(db_span.data["mongo"]["service"], "%s:%s" % (testenv['mongodb_host'], testenv['mongodb_port']))
        self.assertEqual(db_span.data["mongo"]["namespace"], "test.records")
        self.assertEqual(db_span.data["mongo"]["command"], "insert")

        assert_is_none(db_span.data["mongo"]["filter"])

    def test_successful_update_query(self):
        with tracer.start_active_span("test"):
            self.conn.test.records.update_one({"type": "string"}, {"$set": {"type": "int"}})

        assert_is_none(tracer.active_span)

        spans = self.recorder.queued_spans()
        self.assertEqual(len(spans), 2)

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        assert_is_none(db_span.ec)

        self.assertEqual(db_span.n, "mongo")
        self.assertEqual(db_span.data["mongo"]["service"], "%s:%s" % (testenv['mongodb_host'], testenv['mongodb_port']))
        self.assertEqual(db_span.data["mongo"]["namespace"], "test.records")
        self.assertEqual(db_span.data["mongo"]["command"], "update")

        assert_is_none(db_span.data["mongo"]["filter"])
        assert_is_not_none(db_span.data["mongo"]["json"])

        payload = json.loads(db_span.data["mongo"]["json"])
        assert_true({
                        "q": {"type": "string"},
                        "u": {"$set": {"type": "int"}},
                        "multi": False,
                        "upsert": False
                    } in payload, db_span.data["mongo"]["json"])

    def test_successful_delete_query(self):
        with tracer.start_active_span("test"):
            self.conn.test.records.delete_one(filter={"type": "string"})

        assert_is_none(tracer.active_span)

        spans = self.recorder.queued_spans()
        self.assertEqual(len(spans), 2)

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        assert_is_none(db_span.ec)

        self.assertEqual(db_span.n, "mongo")
        self.assertEqual(db_span.data["mongo"]["service"], "%s:%s" % (testenv['mongodb_host'], testenv['mongodb_port']))
        self.assertEqual(db_span.data["mongo"]["namespace"], "test.records")
        self.assertEqual(db_span.data["mongo"]["command"], "delete")

        assert_is_none(db_span.data["mongo"]["filter"])
        assert_is_not_none(db_span.data["mongo"]["json"])

        payload = json.loads(db_span.data["mongo"]["json"])
        assert_true({"q": {"type": "string"}, "limit": 1} in payload, db_span.data["mongo"]["json"])

    def test_successful_aggregate_query(self):
        with tracer.start_active_span("test"):
            self.conn.test.records.count_documents({"type": "string"})

        assert_is_none(tracer.active_span)

        spans = self.recorder.queued_spans()
        self.assertEqual(len(spans), 2)

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        assert_is_none(db_span.ec)

        self.assertEqual(db_span.n, "mongo")
        self.assertEqual(db_span.data["mongo"]["service"], "%s:%s" % (testenv['mongodb_host'], testenv['mongodb_port']))
        self.assertEqual(db_span.data["mongo"]["namespace"], "test.records")
        self.assertEqual(db_span.data["mongo"]["command"], "aggregate")

        assert_is_none(db_span.data["mongo"]["filter"])
        assert_is_not_none(db_span.data["mongo"]["json"])

        payload = json.loads(db_span.data["mongo"]["json"])
        assert_true({"$match": {"type": "string"}} in payload, db_span.data["mongo"]["json"])

    def test_successful_map_reduce_query(self):
        mapper = "function () { this.tags.forEach(function(z) { emit(z, 1); }); }"
        reducer = "function (key, values) { return len(values); }"

        with tracer.start_active_span("test"):
            self.conn.test.records.map_reduce(bson.code.Code(mapper), bson.code.Code(reducer), "results",
                                              query={"x": {"$lt": 2}})

        assert_is_none(tracer.active_span)

        spans = self.recorder.queued_spans()
        self.assertEqual(len(spans), 2)

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        assert_is_none(db_span.ec)

        self.assertEqual(db_span.n, "mongo")
        self.assertEqual(db_span.data["mongo"]["service"], "%s:%s" % (testenv['mongodb_host'], testenv['mongodb_port']))
        self.assertEqual(db_span.data["mongo"]["namespace"], "test.records")
        self.assertEqual(db_span.data["mongo"]["command"].lower(),
                         "mapreduce")  # mapreduce command was renamed to mapReduce in pymongo 3.9.0

        self.assertEqual(db_span.data["mongo"]["filter"], '{"x": {"$lt": 2}}')
        assert_is_not_none(db_span.data["mongo"]["json"])

        payload = json.loads(db_span.data["mongo"]["json"])
        self.assertEqual(payload["map"], {"$code": mapper}, db_span.data["mongo"]["json"])
        self.assertEqual(payload["reduce"], {"$code": reducer}, db_span.data["mongo"]["json"])

    def test_successful_mutiple_queries(self):
        with tracer.start_active_span("test"):
            self.conn.test.records.bulk_write([pymongo.InsertOne({"type": "string"}),
                                               pymongo.UpdateOne({"type": "string"}, {"$set": {"type": "int"}}),
                                               pymongo.DeleteOne({"type": "string"})])

        assert_is_none(tracer.active_span)

        spans = self.recorder.queued_spans()
        self.assertEqual(len(spans), 4)

        test_span = spans.pop()

        seen_span_ids = set()
        commands = []
        for span in spans:
            self.assertEqual(test_span.t, span.t)
            self.assertEqual(span.p, test_span.s)

            # check if all spans got a unique id
            assert_false(span.s in seen_span_ids)

            seen_span_ids.add(span.s)
            commands.append(span.data["mongo"]["command"])

        # ensure spans are ordered the same way as commands
        assert_list_equal(commands, ["insert", "update", "delete"])




# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.routing import Route, Mount, WebSocketRoute
from starlette.staticfiles import StaticFiles

import os 
dir_path = os.path.dirname(os.path.realpath(__file__))

def homepage(request):
    return PlainTextResponse('Hello, world!')

def user(request):
    user_id = request.path_params['user_id']
    return PlainTextResponse('Hello, user id %s!' % user_id)

async def websocket_endpoint(websocket):
    await websocket.accept()
    await websocket.send_text('Hello, websocket!')
    await websocket.close()

def startup():
    print('Ready to go')


routes = [
    Route('/', homepage),
    Route('/users/{user_id}', user),
    WebSocketRoute('/ws', websocket_endpoint),
    Mount('/static', StaticFiles(directory=dir_path + "/static")),
]

starlette_server = Starlette(debug=True, routes=routes, on_startup=[startup])


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import os
import time
import pytest
import random
import unittest

from instana.singletons import tracer
from ..helpers import testenv, get_first_span_by_name

from cassandra.cluster import Cluster
from cassandra import ConsistencyLevel
from cassandra.query import SimpleStatement

cluster = Cluster([testenv['cassandra_host']], load_balancing_policy=None)
session = cluster.connect()

session.execute("CREATE KEYSPACE IF NOT EXISTS instana_tests WITH replication = {'class':'SimpleStrategy', 'replication_factor':1};")
session.set_keyspace('instana_tests')
session.execute("CREATE TABLE IF NOT EXISTS users("
                "id int PRIMARY KEY,"
                "name text,"
                "age text,"
                "email varint,"
                "phone varint"
                ");")


@pytest.mark.skipif("CASSANDRA_TEST" not in os.environ, reason="")
class TestCassandra(unittest.TestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = tracer.recorder
        self.recorder.clear_spans()

    def tearDown(self):
        """ Do nothing for now """
        return None

    def test_untraced_execute(self):
        res = session.execute('SELECT name, age, email FROM users')

        self.assertIsNotNone(res)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        self.assertEqual(0, len(spans))

    def test_untraced_execute_error(self):
        res = None
        try:
            res = session.execute('Not a valid query')
        except:
            pass

        self.assertIsNone(res)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        self.assertEqual(0, len(spans))

    def test_execute(self):
        res = None
        with tracer.start_active_span('test'):
            res = session.execute('SELECT name, age, email FROM users')

        self.assertIsNotNone(res)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        test_span = get_first_span_by_name(spans, 'sdk')
        self.assertIsNotNone(test_span)
        self.assertEqual(test_span.data["sdk"]["name"], 'test')

        cspan = get_first_span_by_name(spans, 'cassandra')
        self.assertIsNotNone(cspan)

        # Same traceId and parent relationship
        self.assertEqual(test_span.t, cspan.t)
        self.assertEqual(cspan.p, test_span.s)

        self.assertIsNotNone(cspan.stack)
        self.assertIsNone(cspan.ec)

        self.assertEqual(cspan.data["cassandra"]["cluster"], 'Test Cluster')
        self.assertEqual(cspan.data["cassandra"]["query"], 'SELECT name, age, email FROM users')
        self.assertEqual(cspan.data["cassandra"]["keyspace"], 'instana_tests')
        self.assertIsNone(cspan.data["cassandra"]["achievedConsistency"])
        self.assertIsNotNone(cspan.data["cassandra"]["triedHosts"])
        self.assertIsNone(cspan.data["cassandra"]["error"])

    def test_execute_async(self):
        res = None
        with tracer.start_active_span('test'):
            res = session.execute_async('SELECT name, age, email FROM users').result()

        self.assertIsNotNone(res)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        test_span = get_first_span_by_name(spans, 'sdk')
        self.assertIsNotNone(test_span)
        self.assertEqual(test_span.data["sdk"]["name"], 'test')

        cspan = get_first_span_by_name(spans, 'cassandra')
        self.assertIsNotNone(cspan)

        # Same traceId and parent relationship
        self.assertEqual(test_span.t, cspan.t)
        self.assertEqual(cspan.p, test_span.s)

        self.assertIsNotNone(cspan.stack)
        self.assertIsNone(cspan.ec)

        self.assertEqual(cspan.data["cassandra"]["cluster"], 'Test Cluster')
        self.assertEqual(cspan.data["cassandra"]["query"], 'SELECT name, age, email FROM users')
        self.assertEqual(cspan.data["cassandra"]["keyspace"], 'instana_tests')
        self.assertIsNone(cspan.data["cassandra"]["achievedConsistency"])
        self.assertIsNotNone(cspan.data["cassandra"]["triedHosts"])
        self.assertIsNone(cspan.data["cassandra"]["error"])

    def test_simple_statement(self):
        res = None
        with tracer.start_active_span('test'):
            query = SimpleStatement(
                'SELECT name, age, email FROM users',
                is_idempotent=True
            )
            res = session.execute(query)

        self.assertIsNotNone(res)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        test_span = get_first_span_by_name(spans, 'sdk')
        self.assertIsNotNone(test_span)
        self.assertEqual(test_span.data["sdk"]["name"], 'test')

        cspan = get_first_span_by_name(spans, 'cassandra')
        self.assertIsNotNone(cspan)

        # Same traceId and parent relationship
        self.assertEqual(test_span.t, cspan.t)
        self.assertEqual(cspan.p, test_span.s)

        self.assertIsNotNone(cspan.stack)
        self.assertIsNone(cspan.ec)

        self.assertEqual(cspan.data["cassandra"]["cluster"], 'Test Cluster')
        self.assertEqual(cspan.data["cassandra"]["query"], 'SELECT name, age, email FROM users')
        self.assertEqual(cspan.data["cassandra"]["keyspace"], 'instana_tests')
        self.assertIsNone(cspan.data["cassandra"]["achievedConsistency"])
        self.assertIsNotNone(cspan.data["cassandra"]["triedHosts"])
        self.assertIsNone(cspan.data["cassandra"]["error"])

    def test_execute_error(self):
        res = None

        try:
            with tracer.start_active_span('test'):
                res = session.execute('Not a real query')
        except:
            pass

        self.assertIsNone(res)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        test_span = get_first_span_by_name(spans, 'sdk')
        self.assertIsNotNone(test_span)
        self.assertEqual(test_span.data["sdk"]["name"], 'test')

        cspan = get_first_span_by_name(spans, 'cassandra')
        self.assertIsNotNone(cspan)

        # Same traceId and parent relationship
        self.assertEqual(test_span.t, cspan.t)
        self.assertEqual(cspan.p, test_span.s)

        self.assertIsNotNone(cspan.stack)
        self.assertEqual(cspan.ec, 1)

        self.assertEqual(cspan.data["cassandra"]["cluster"], 'Test Cluster')
        self.assertEqual(cspan.data["cassandra"]["query"], 'Not a real query')
        self.assertEqual(cspan.data["cassandra"]["keyspace"], 'instana_tests')
        self.assertIsNone(cspan.data["cassandra"]["achievedConsistency"])
        self.assertIsNotNone(cspan.data["cassandra"]["triedHosts"])
        self.assertIsNotNone(cspan.data["cassandra"]["error"])

    def test_prepared_statement(self):
        prepared = None
        result = None

        with tracer.start_active_span('test'):
            prepared = session.prepare('INSERT INTO users (id, name, age) VALUES (?, ?, ?)')
            prepared.consistency_level = ConsistencyLevel.QUORUM
            result = session.execute(prepared, (random.randint(0, 1000000), "joe", "17"))

        self.assertIsNotNone(prepared)
        self.assertIsNotNone(result)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        test_span = get_first_span_by_name(spans, 'sdk')
        self.assertIsNotNone(test_span)
        self.assertEqual(test_span.data["sdk"]["name"], 'test')

        cspan = get_first_span_by_name(spans, 'cassandra')
        self.assertIsNotNone(cspan)

        # Same traceId and parent relationship
        self.assertEqual(test_span.t, cspan.t)
        self.assertEqual(cspan.p, test_span.s)

        self.assertIsNotNone(cspan.stack)
        self.assertIsNone(cspan.ec)

        self.assertEqual(cspan.data["cassandra"]["cluster"], 'Test Cluster')
        self.assertEqual(cspan.data["cassandra"]["query"], 'INSERT INTO users (id, name, age) VALUES (?, ?, ?)')
        self.assertEqual(cspan.data["cassandra"]["keyspace"], 'instana_tests')
        self.assertEqual(cspan.data["cassandra"]["achievedConsistency"], "QUORUM")
        self.assertIsNotNone(cspan.data["cassandra"]["triedHosts"])
        self.assertIsNone(cspan.data["cassandra"]["error"])



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from __future__ import absolute_import

import os
import sys
import threading
import time
import pytest

import six
import unittest

from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient
from google.api_core.exceptions import AlreadyExists
from google.cloud.pubsub_v1.publisher import exceptions
from instana.singletons import tracer
from tests.test_utils import _TraceContextMixin

# Use PubSub Emulator exposed at :8432
os.environ["PUBSUB_EMULATOR_HOST"] = "localhost:8432"


@pytest.mark.skipif(sys.version_info[0] < 3,
                    reason="google-cloud-pubsub has dropped support for Python 2")
class TestPubSubPublish(unittest.TestCase, _TraceContextMixin):
    @classmethod
    def setUpClass(cls):
        cls.publisher = PublisherClient()

    def setUp(self):
        self.recorder = tracer.recorder
        self.recorder.clear_spans()

        self.project_id = 'test-project'
        self.topic_name = 'test-topic'

        # setup topic_path & topic
        self.topic_path = self.publisher.topic_path(self.project_id, self.topic_name)
        try:
            self.publisher.create_topic(request={"name": self.topic_path})
        except AlreadyExists:
            self.publisher.delete_topic(request={"topic": self.topic_path})
            self.publisher.create_topic(request={"name": self.topic_path})

    def tearDown(self):
        self.publisher.delete_topic(request={"topic": self.topic_path})

    def test_publish(self):
        # publish a single message
        with tracer.start_active_span('test'):
            future = self.publisher.publish(self.topic_path,
                                            b'Test Message',
                                            origin="instana")
        time.sleep(2.0)  # for sanity
        result = future.result()
        assert isinstance(result, six.string_types)

        spans = self.recorder.queued_spans()
        gcps_span, test_span = spans[0], spans[1]

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)
        self.assertEqual('gcps', gcps_span.n)
        self.assertEqual(2, gcps_span.k)  # EXIT

        self.assertEqual('publish', gcps_span.data['gcps']['op'])
        self.assertEqual(self.topic_name, gcps_span.data['gcps']['top'])

        # Trace Context Propagation
        self.assertTraceContextPropagated(test_span, gcps_span)

        # Error logging
        self.assertErrorLogging(spans)


class AckCallback(object):
    def __init__(self):
        self.calls = 0
        self.lock = threading.Lock()

    def __call__(self, message):
        message.ack()
        # Only increment the number of calls **after** finishing.
        with self.lock:
            self.calls += 1


@pytest.mark.skipif(sys.version_info[0] < 3,
                    reason="google-cloud-pubsub has dropped support for Python 2")
class TestPubSubSubscribe(unittest.TestCase, _TraceContextMixin):
    @classmethod
    def setUpClass(cls):
        cls.publisher = PublisherClient()
        cls.subscriber = SubscriberClient()

    def setUp(self):

        self.recorder = tracer.recorder
        self.recorder.clear_spans()

        self.project_id = 'test-project'
        self.topic_name = 'test-topic'
        self.subscription_name = 'test-subscription'

        # setup topic_path & topic
        self.topic_path = self.publisher.topic_path(self.project_id, self.topic_name)
        try:
            self.publisher.create_topic(request={"name": self.topic_path})
        except AlreadyExists:
            self.publisher.delete_topic(request={"topic": self.topic_path})
            self.publisher.create_topic(request={"name": self.topic_path})

        # setup subscription path & attach subscription
        self.subscription_path = self.subscriber.subscription_path(
            self.project_id, self.subscription_name)
        try:
            self.subscriber.create_subscription(
                request={"name": self.subscription_path, "topic": self.topic_path}
            )
        except AlreadyExists:
            self.subscriber.delete_subscription(request={"subscription": self.subscription_path})
            self.subscriber.create_subscription(
                request={"name": self.subscription_path, "topic": self.topic_path}
            )

    def tearDown(self):
        self.publisher.delete_topic(request={"topic": self.topic_path})
        self.subscriber.delete_subscription(request={"subscription": self.subscription_path})

    def test_subscribe(self):

        with tracer.start_active_span('test'):
            # Publish a message
            future = self.publisher.publish(self.topic_path,
                                            b"Test Message to PubSub",
                                            origin="instana")
            self.assertIsInstance(future.result(), six.string_types)

            time.sleep(2.0)  # for sanity

            # Subscribe to the subscription
            callback_handler = AckCallback()
            future = self.subscriber.subscribe(self.subscription_path, callback_handler)
            timeout = 2.0
            try:
                future.result(timeout)
            except exceptions.TimeoutError:
                future.cancel()

        spans = self.recorder.queued_spans()

        producer_span = spans[0]
        consumer_span = spans[1]
        test_span = spans[2]

        self.assertEqual(3, len(spans))
        self.assertIsNone(tracer.active_span)
        self.assertEqual('publish', producer_span.data['gcps']['op'])
        self.assertEqual('consume', consumer_span.data['gcps']['op'])
        self.assertEqual(self.topic_name, producer_span.data['gcps']['top'])
        self.assertEqual(self.subscription_name, consumer_span.data['gcps']['sub'])

        self.assertEqual(2, producer_span.k)  # EXIT
        self.assertEqual(1, consumer_span.k)  # ENTRY

        # Trace Context Propagation
        self.assertTraceContextPropagated(producer_span, consumer_span)
        self.assertTraceContextPropagated(test_span, producer_span)

        # Error logging
        self.assertErrorLogging(spans)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from instana.collector.helpers.process import ProcessHelper
from instana.log import logger


class GCRProcessHelper(ProcessHelper):
    """ Helper class to extend the generic process helper class with the corresponding Google Cloud Run attributes """

    def collect_metrics(self, **kwargs):
        plugin_data = dict()
        try:
            plugin_data = super(GCRProcessHelper, self).collect_metrics(**kwargs)
            plugin_data["data"]["containerType"] = "gcpCloudRunInstance"
            plugin_data["data"]["container"] = self.collector.get_instance_id()
            plugin_data["data"]["com.instana.plugin.host.name"] = "gcp:cloud-run:revision:{revision}".format(
                revision=self.collector.revision)
        except Exception:
            logger.debug("GCRProcessHelper.collect_metrics: ", exc_info=True)
        return [plugin_data]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

from __future__ import absolute_import

try:
    import flask
    from flask.signals import signals_available

    # `signals_available` indicates whether the Flask process is running with or without blinker support:
    # https://pypi.org/project/blinker/
    #
    # Blinker support is preferred but we do the best we can when it's not available.
    #

    from . import common

    if signals_available is True:
        import instana.instrumentation.flask.with_blinker
    else:
        import instana.instrumentation.flask.vanilla
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021


from sanic.views import HTTPMethodView
from sanic.response import text


class NameView(HTTPMethodView):

    def get(self, request, name):
        return text("Hello {}".format(name))





# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

import os
import sys
import time
import threading

if 'GEVENT_TEST' not in os.environ and 'CASSANDRA_TEST' not in os.environ and sys.version_info >= (3, 5, 3):
    # Background RPC application
    #
    # Spawn the background RPC app that the tests will throw
    # requests at.
    import tests.apps.grpc_server
    from .stan_server import StanServicer
    stan_servicer = StanServicer()
    rpc_server_thread = threading.Thread(target=stan_servicer.start_server)
    rpc_server_thread.daemon = True
    rpc_server_thread.name = "Background RPC app"
    print("Starting background RPC app...")
    rpc_server_thread.start()
    time.sleep(1)


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
AWS Lambda Collector: Manages the periodic collection of metrics & snapshot data
"""
from ..log import logger
from .base import BaseCollector
from ..util import DictionaryOfStan
from ..util.aws import normalize_aws_lambda_arn


class AWSLambdaCollector(BaseCollector):
    """ Collector for AWS Lambda """
    def __init__(self, agent):
        super(AWSLambdaCollector, self).__init__(agent)
        logger.debug("Loading AWS Lambda Collector")
        self.context = None
        self.event = None
        self._fq_arn = None

        # How often to report data
        self.report_interval = 5

        self.snapshot_data = DictionaryOfStan()
        self.snapshot_data_sent = False

    def collect_snapshot(self, event, context):
        self.context = context
        self.event = event

        try:
            plugin_data = dict()
            plugin_data["name"] = "com.instana.plugin.aws.lambda"
            plugin_data["entityId"] = self.get_fq_arn()
            self.snapshot_data["plugins"] = [plugin_data]
        except Exception:
            logger.debug("collect_snapshot error", exc_info=True)
        return self.snapshot_data

    def should_send_snapshot_data(self):
        return self.snapshot_data and self.snapshot_data_sent is False

    def prepare_payload(self):
        payload = DictionaryOfStan()
        payload["spans"] = None
        payload["metrics"] = None

        if not self.span_queue.empty():
            payload["spans"] = self.queued_spans()

        if self.should_send_snapshot_data():
            payload["metrics"] = self.snapshot_data
            self.snapshot_data_sent = True

        return payload

    def get_fq_arn(self):
        if self._fq_arn is not None:
            return self._fq_arn

        if self.context is None:
            logger.debug("Attempt to get qualified ARN before the context object is available")
            return ''

        self._fq_arn = normalize_aws_lambda_arn(self.context)
        return self._fq_arn



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import pytest
import unittest
import tests.apps.soap_app
from ..helpers import testenv
from suds.client import Client
from instana.singletons import tracer


@pytest.mark.skip(reason="Unstable tests")
class TestSudsJurko(unittest.TestCase):
    def setup_class(self):
        """ Clear all spans before a test run """
        self.client = Client(testenv["soap_server"] + '/?wsdl', cache=None)
        self.recorder = tracer.recorder

    def setup_method(self):
        self.recorder.clear_spans()
        tracer.cur_ctx = None

    def test_vanilla_request(self):
        response = self.client.service.ask_question(u'Why u like dat?', 5)

        self.assertEqual(1, len(response))
        self.assertEqual(1, len(response[0]))
        assert(type(response[0]) is list)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

    def test_basic_request(self):
        with tracer.start_active_span('test'):
            response = self.client.service.ask_question(u'Why u like dat?', 5)

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))
        wsgi_span = spans[0]
        soap_span = spans[1]
        test_span = spans[2]

        self.assertEqual(1, len(response))
        self.assertEqual(1, len(response[0]))
        assert(type(response[0]) is list)

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, soap_span.t)
        self.assertEqual(soap_span.p, test_span.s)
        self.assertEqual(wsgi_span.t, soap_span.t)
        self.assertEqual(wsgi_span.p, soap_span.s)

        self.assertEqual(None, soap_span.ec)

        self.assertEqual('ask_question', soap_span.data["soap"]["action"])
        self.assertEqual(testenv["soap_server"] + '/', soap_span.data["http"]["url"])

    def test_server_exception(self):
        response = None
        with tracer.start_active_span('test'):
            try:
                response = self.client.service.server_exception()
            except Exception:
                pass

        spans = self.recorder.queued_spans()
        self.assertEqual(5, len(spans))

        log_span1 = spans[0]
        wsgi_span = spans[1]
        log_span2 = spans[2]
        soap_span = spans[3]
        test_span = spans[4]

        self.assertEqual(None, response)
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, soap_span.t)
        self.assertEqual(soap_span.p, test_span.s)
        self.assertEqual(wsgi_span.t, soap_span.t)
        self.assertEqual(wsgi_span.p, soap_span.s)

        self.assertEqual(1, soap_span.ec)
        self.assertEqual(u"Server raised fault: 'Internal Error'", soap_span.data["http"]["error"])
        self.assertEqual('server_exception', soap_span.data["soap"]["action"])
        self.assertEqual(testenv["soap_server"] + '/', soap_span.data["http"]["url"])

    def test_server_fault(self):
        response = None
        with tracer.start_active_span('test'):
            try:
                response = self.client.service.server_fault()
            except Exception:
                pass

        spans = self.recorder.queued_spans()
        self.assertEqual(5, len(spans))
        log_span1 = spans[0]
        wsgi_span = spans[1]
        log_span2 = spans[2]
        soap_span = spans[3]
        test_span = spans[4]

        self.assertEqual(None, response)
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, soap_span.t)
        self.assertEqual(soap_span.p, test_span.s)
        self.assertEqual(wsgi_span.t, soap_span.t)
        self.assertEqual(wsgi_span.p, soap_span.s)

        self.assertEqual(1, soap_span.ec)
        self.assertEqual(u"Server raised fault: 'Server side fault example.'", soap_span.data["http"]["error"])
        self.assertEqual('server_fault', soap_span.data["soap"]["action"])
        self.assertEqual(testenv["soap_server"] + '/', soap_span.data["http"]["url"])

    def test_client_fault(self):
        response = None
        with tracer.start_active_span('test'):
            try:
                response = self.client.service.client_fault()
            except Exception:
                pass

        spans = self.recorder.queued_spans()
        self.assertEqual(5, len(spans))

        log_span1 = spans[0]
        wsgi_span = spans[1]
        log_span2 = spans[2]
        soap_span = spans[3]
        test_span = spans[4]

        self.assertEqual(None, response)
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, soap_span.t)
        self.assertEqual(soap_span.p, test_span.s)
        self.assertEqual(wsgi_span.t, soap_span.t)
        self.assertEqual(wsgi_span.p, soap_span.s)

        self.assertEqual(1, soap_span.ec)
        self.assertEqual(u"Server raised fault: 'Client side fault example'", soap_span.data["http"]["error"])
        self.assertEqual('client_fault', soap_span.data["soap"]["action"])
        self.assertEqual(testenv["soap_server"] + '/', soap_span.data["http"]["url"])



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
The in-process Instana agent (for host based processes) that manages
monitoring state and reporting that data.
"""
from __future__ import absolute_import

import os
import json
from datetime import datetime

import urllib3
import requests

from ..log import logger
from .base import BaseAgent
from ..fsm import TheMachine
from ..version import VERSION
from ..options import StandardOptions
from ..collector.host import HostCollector
from ..util import to_json
from ..util.runtime import get_py_source


class AnnounceData(object):
    """ The Announce Payload """
    pid = 0
    agentUuid = ""

    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class HostAgent(BaseAgent):
    """
    The Agent class is the central controlling entity for the Instana Python language sensor.  The key
    parts it handles are the announce state and the collection and reporting of metrics and spans to the
    Instana Host agent.
    """
    AGENT_DISCOVERY_PATH = "com.instana.plugin.python.discovery"
    AGENT_DATA_PATH = "com.instana.plugin.python.%d"
    AGENT_HEADER = "Instana Agent"

    def __init__(self):
        super(HostAgent, self).__init__()

        self.announce_data = None
        self.machine = None
        self.last_seen = None
        self.last_fork_check = None
        self._boot_pid = os.getpid()
        self.options = StandardOptions()

        # Update log level from what Options detected
        self.update_log_level()

        logger.info("Stan is on the scene.  Starting Instana instrumentation version: %s", VERSION)

        self.collector = HostCollector(self)
        self.machine = TheMachine(self)

    def start(self):
        """
        Starts the agent and required threads

        This method is called after a successful announce.  See fsm.py
        """
        logger.debug("Starting Host Collector")
        self.collector.start()

    def handle_fork(self):
        """
        Forks happen.  Here we handle them.
        """
        # Reset the Agent
        self.reset()

    def reset(self):
        """
        This will reset the agent to a fresh unannounced state.
        :return: None
        """
        self.last_seen = None
        self.announce_data = None
        self.collector.shutdown(report_final=False)

        # Will schedule a restart of the announce cycle in the future
        self.machine.reset()

    def is_timed_out(self):
        """
        If we haven't heard from the Instana host agent in 60 seconds, this
        method will return True.
        @return: Boolean
        """
        if self.last_seen and self.can_send:
            diff = datetime.now() - self.last_seen
            if diff.seconds > 60:
                return True
        return False

    def can_send(self):
        """
        Are we in a state where we can send data?
        @return: Boolean
        """
        if "INSTANA_TEST" in os.environ:
            return True

        # Watch for pid change (fork)
        self.last_fork_check = datetime.now()
        current_pid = os.getpid()
        if self._boot_pid != current_pid:
            self._boot_pid = current_pid
            logger.debug("Fork detected; Handling like a pro...")
            self.handle_fork()
            return False

        if self.machine.fsm.current in ["wait4init", "good2go"]:
            return True

        return False

    def set_from(self, json_string):
        """
        Sets the source identifiers given to use by the Instana Host agent.
        @param json_string: source identifiers
        @return: None
        """
        if isinstance(json_string, bytes):
            raw_json = json_string.decode("UTF-8")
        else:
            raw_json = json_string

        res_data = json.loads(raw_json)

        if "secrets" in res_data:
            self.options.secrets_matcher = res_data['secrets']['matcher']
            self.options.secrets_list = res_data['secrets']['list']

        if "extraHeaders" in res_data:
            if self.options.extra_http_headers is None:
                self.options.extra_http_headers = res_data['extraHeaders']
            else:
                self.options.extra_http_headers.extend(res_data['extraHeaders'])
            logger.info("Will also capture these custom headers: %s", self.options.extra_http_headers)

        self.announce_data = AnnounceData(pid=res_data['pid'], agentUuid=res_data['agentUuid'])

    def get_from_structure(self):
        """
        Retrieves the From data that is reported alongside monitoring data.
        @return: dict()
        """
        return {'e': self.announce_data.pid, 'h': self.announce_data.agentUuid}

    def is_agent_listening(self, host, port):
        """
        Check if the Instana Agent is listening on <host> and <port>.
        @return: Boolean
        """
        result = False
        try:
            url = "http://%s:%s/" % (host, port)
            response = self.client.get(url, timeout=0.8)

            server_header = response.headers["Server"]
            if server_header == self.AGENT_HEADER:
                logger.debug("Instana host agent found on %s:%d", host, port)
                result = True
            else:
                logger.debug("...something is listening on %s:%d but it's not the Instana Host Agent: %s",
                             host, port, server_header)
        except Exception:
            logger.debug("Instana Host Agent not found on %s:%d", host, port)
        return result

    def announce(self, discovery):
        """
        With the passed in Discovery class, attempt to announce to the host agent.
        """
        response = None
        try:
            url = self.__discovery_url()
            response = self.client.put(url,
                                       data=to_json(discovery),
                                       headers={"Content-Type": "application/json"},
                                       timeout=0.8)

            if 200 <= response.status_code <= 204:
                self.last_seen = datetime.now()
        except Exception as exc:
            logger.debug("announce: connection error (%s)", type(exc))
        return response

    def log_message_to_host_agent(self, message):
        """
        Log a message to the discovered host agent
        """
        response = None
        try:
            payload = dict()
            payload["m"] = message

            url = self.__agent_logger_url()
            response = self.client.post(url,
                                       data=to_json(payload),
                                       headers={"Content-Type": "application/json",
                                                "X-Log-Level": "INFO"},
                                       timeout=0.8)

            if 200 <= response.status_code <= 204:
                self.last_seen = datetime.now()
        except Exception as exc:
            logger.debug("agent logging: connection error (%s)", type(exc))

    def is_agent_ready(self):
        """
        Used after making a successful announce to test when the agent is ready to accept data.
        """
        ready = False
        try:
            response = self.client.head(self.__data_url(), timeout=0.8)

            if response.status_code == 200:
                ready = True
        except Exception as exc:
            logger.debug("is_agent_ready: connection error (%s)", type(exc))
        return ready

    def report_data_payload(self, payload):
        """
        Used to report collection payload to the host agent.  This can be metrics, spans and snapshot data.
        """
        response = None
        try:
            # Report spans (if any)
            span_count = len(payload['spans'])
            if span_count > 0:
                logger.debug("Reporting %d spans", span_count)
                response = self.client.post(self.__traces_url(),
                                            data=to_json(payload['spans']),
                                            headers={"Content-Type": "application/json"},
                                            timeout=0.8)

            if response is not None and 200 <= response.status_code <= 204:
                self.last_seen = datetime.now()

            # Report profiles (if any)
            profile_count = len(payload['profiles'])
            if profile_count > 0:
                logger.debug("Reporting %d profiles", profile_count)
                response = self.client.post(self.__profiles_url(),
                                            data=to_json(payload['profiles']),
                                            headers={"Content-Type": "application/json"},
                                            timeout=0.8)

            if response is not None and 200 <= response.status_code <= 204:
                self.last_seen = datetime.now()

            # Report metrics
            metric_bundle = payload["metrics"]["plugins"][0]["data"]
            response = self.client.post(self.__data_url(),
                                        data=to_json(metric_bundle),
                                        headers={"Content-Type": "application/json"},
                                        timeout=0.8)

            if response is not None and 200 <= response.status_code <= 204:
                self.last_seen = datetime.now()

                if response.status_code == 200 and len(response.content) > 2:
                    # The host agent returned something indicating that is has a request for us that we
                    # need to process.
                    self.handle_agent_tasks(json.loads(response.content)[0])
        except requests.exceptions.ConnectionError:
            pass
        except urllib3.exceptions.MaxRetryError:
            pass
        except Exception as exc:
            logger.debug("report_data_payload: Instana host agent connection error (%s)", type(exc), exc_info=True)
        return response

    def handle_agent_tasks(self, task):
        """
        When request(s) are received by the host agent, it is sent here
        for handling & processing.
        """
        logger.debug("Received agent request with messageId: %s", task["messageId"])
        if "action" in task:
            if task["action"] == "python.source":
                payload = get_py_source(task["args"]["file"])
            else:
                message = "Unrecognized action: %s. An newer Instana package may be required " \
                          "for this. Current version: %s" % (task["action"], VERSION)
                payload = {"error": message}
        else:
            payload = {"error": "Instana Python: No action specified in request."}

        self.__task_response(task["messageId"], payload)


    def diagnostics(self):
        """
        Helper function to dump out state.
        """
        try:
            import threading
            dt_format = "%Y-%m-%d %H:%M:%S"

            logger.warning("====> Instana Python Language Agent Diagnostics <====")

            logger.warning("----> Agent <----")
            logger.warning("is_agent_ready: %s", self.is_agent_ready())
            logger.warning("is_timed_out: %s", self.is_timed_out())
            if self.last_seen is None:
                logger.warning("last_seen: None")
            else:
                logger.warning("last_seen: %s", self.last_seen.strftime(dt_format))

            if self.announce_data is not None:
                logger.warning("announce_data: %s", self.announce_data.__dict__)
            else:
                logger.warning("announce_data: None")

            logger.warning("Options: %s", self.options.__dict__)

            logger.warning("----> StateMachine <----")
            logger.warning("State: %s", self.machine.fsm.current)

            logger.warning("----> Collector <----")
            logger.warning("Collector: %s", self.collector)
            logger.warning("is_collector_thread_running?: %s", self.collector.is_reporting_thread_running())
            logger.warning("background_report_lock.locked?: %s", self.collector.background_report_lock.locked())
            logger.warning("ready_to_start: %s", self.collector.ready_to_start)
            logger.warning("reporting_thread: %s", self.collector.reporting_thread)
            logger.warning("report_interval: %s", self.collector.report_interval)
            logger.warning("should_send_snapshot_data: %s", self.collector.should_send_snapshot_data())
            logger.warning("spans in queue: %s", self.collector.span_queue.qsize())
            logger.warning("thread_shutdown is_set: %s", self.collector.thread_shutdown.is_set())

            logger.warning("----> Threads <----")
            logger.warning("Threads: %s", threading.enumerate())
        except Exception:
            logger.warning("Non-fatal diagnostics exception: ", exc_info=True)

    def __task_response(self, message_id, data):
        """
        When the host agent passes us a task and we do it, this function is used to
        respond with the results of the task.
        """
        response = None
        try:
            payload = json.dumps(data)

            logger.debug("Task response is %s: %s", self.__response_url(message_id), payload)

            response = self.client.post(self.__response_url(message_id),
                                        data=payload,
                                        headers={"Content-Type": "application/json"},
                                        timeout=0.8)
        except Exception as exc:
            logger.debug("__task_response: Instana host agent connection error (%s)", type(exc))
        return response

    def __discovery_url(self):
        """
        URL for announcing to the host agent
        """
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, self.AGENT_DISCOVERY_PATH)

    def __data_url(self):
        """
        URL for posting metrics to the host agent.  Only valid when announced.
        """
        path = self.AGENT_DATA_PATH % self.announce_data.pid
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, path)

    def __traces_url(self):
        """
        URL for posting traces to the host agent.  Only valid when announced.
        """
        path = "com.instana.plugin.python/traces.%d" % self.announce_data.pid
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, path)

    def __profiles_url(self):
        """
        URL for posting profiles to the host agent.  Only valid when announced.
        """
        path = "com.instana.plugin.python/profiles.%d" % self.announce_data.pid
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, path)

    def __response_url(self, message_id):
        """
        URL for responding to agent requests.
        """
        path = "com.instana.plugin.python/response.%d?messageId=%s" % (int(self.announce_data.pid), message_id)
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, path)

    def __agent_logger_url(self):
        """
        URL for logging messages to the discovered host agent.
        """
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, "com.instana.agent.logger")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from instana.w3c_trace_context.tracestate import Tracestate
import unittest


class TestTracestate(unittest.TestCase):
    def setUp(self):
        self.ts = Tracestate()

    def test_get_instana_ancestor(self):
        tracestate = "congo=t61rcWkgMzE,in=1234d0e0e4736234;1234567890abcdef"
        ia = self.ts.get_instana_ancestor(tracestate)
        self.assertEqual(ia.t, "1234d0e0e4736234")
        self.assertEqual(ia.p, "1234567890abcdef")

    def test_get_instana_ancestor_no_in(self):
        tracestate = "congo=t61rcWkgMzE"
        self.assertIsNone(self.ts.get_instana_ancestor(tracestate))

    def test_get_instana_ancestor_tracestate_None(self):
        tracestate = None
        self.assertIsNone(self.ts.get_instana_ancestor(tracestate))

    def test_update_tracestate(self):
        tracestate = "congo=t61rcWkgMzE"
        in_trace_id = "1234d0e0e4736234"
        in_span_id = "1234567890abcdef"
        expected_tracestate = "in=1234d0e0e4736234;1234567890abcdef,congo=t61rcWkgMzE"
        self.assertEqual(expected_tracestate, self.ts.update_tracestate(tracestate, in_trace_id, in_span_id))

    def test_update_tracestate_None(self):
        tracestate = None
        in_trace_id = "1234d0e0e4736234"
        in_span_id = "1234567890abcdef"
        expected_tracestate = "in=1234d0e0e4736234;1234567890abcdef"
        self.assertEqual(expected_tracestate, self.ts.update_tracestate(tracestate, in_trace_id, in_span_id))

    def test_update_tracestate_more_than_32_members_already(self):
        tracestate = "congo=t61rcWkgMzE,robo=1221213jdfjkdsfjsd,alpha=5889fnjkllllllll," \
                     "beta=aslsdklkljfdshasfaskkfnnnsdsd,gamadeltaepsilonpirpsigma=125646845613675451535445155126666fgsdfdsfjsdfhsdfsdsdsaddfasfdfdsfdsfsd;qwertyuiopasdfghjklzxcvbnm1234567890," \
                     "b=121,c=23344,d=asd,e=ldkfj,f=1212121,g=sadahsda,h=jjhdada,i=eerjrjrr,j=sadsasd,k=44444,l=dadadad," \
                     "m=rrrr,n=3424jdg,p=ffss,q=12,r=3,s=5,t=u5,u=43,v=gj,w=wew,x=23123,y=sdf,z=kasdl,aa=dsdas,ab=res," \
                     "ac=trwa,ad=kll,ae=pds"
        in_trace_id = "1234d0e0e4736234"
        in_span_id = "1234567890abcdef"
        expected_tracestate = "in=1234d0e0e4736234;1234567890abcdef,congo=t61rcWkgMzE,robo=1221213jdfjkdsfjsd," \
                              "alpha=5889fnjkllllllll,beta=aslsdklkljfdshasfaskkfnnnsdsd,b=121,c=23344,d=asd,e=ldkfj," \
                              "f=1212121,g=sadahsda,h=jjhdada,i=eerjrjrr,j=sadsasd,k=44444,l=dadadad,m=rrrr,n=3424jdg," \
                              "p=ffss,q=12,r=3,s=5,t=u5,u=43,v=gj,w=wew,x=23123,y=sdf,z=kasdl,aa=dsdas,ab=res,ac=trwa"
        actual_tracestate = self.ts.update_tracestate(tracestate, in_trace_id, in_span_id)
        self.assertEqual(len(tracestate.split(",")), 34)  # input had 34 list members
        self.assertEqual(len(actual_tracestate.split(",")), 32)  # output has 32 list members, 3 removed and 1 added
        self.assertEqual(expected_tracestate, actual_tracestate)
        self.assertNotIn("gamadeltaepsilonpirpsigma",
                         actual_tracestate)  # member longer than 128 characters gets removed

    def test_update_tracestate_empty_string(self):
        tracestate = ""
        in_trace_id = "1234d0e0e4736234"
        in_span_id = "1234567890abcdef"
        expected_tracestate = "in=1234d0e0e4736234;1234567890abcdef"
        self.assertEqual(expected_tracestate, self.ts.update_tracestate(tracestate, in_trace_id, in_span_id))

    def test_update_tracestate_exception(self):
        tracestate = []
        in_trace_id = "1234d0e0e4736234"
        in_span_id = "1234567890abcdef"
        expected_tracestate = []
        self.assertEqual(expected_tracestate, self.ts.update_tracestate(tracestate, in_trace_id, in_span_id))


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from instana.w3c_trace_context.traceparent import Traceparent
import unittest


class TestTraceparent(unittest.TestCase):
    def setUp(self):
        self.tp = Traceparent()

    def test_validate_valid(self):
        traceparent = "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"
        self.assertEqual(traceparent, self.tp.validate(traceparent))

    def test_validate_invalid_traceparent(self):
        traceparent = "00-4bxxxxx3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"
        self.assertIsNone(self.tp.validate(traceparent))

    def test_validate_traceparent_None(self):
        traceparent = None
        self.assertIsNone(self.tp.validate(traceparent))

    def test_get_traceparent_fields(self):
        traceparent = "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"
        version, trace_id, parent_id, trace_flags = self.tp.get_traceparent_fields(traceparent)
        self.assertEqual(trace_id, "4bf92f3577b34da6a3ce929d0e0e4736")
        self.assertEqual(parent_id, "00f067aa0ba902b7")

    def test_get_traceparent_fields_None_input(self):
        traceparent = None
        version, trace_id, parent_id, trace_flags = self.tp.get_traceparent_fields(traceparent)
        self.assertIsNone(trace_id)
        self.assertIsNone(parent_id)

    def test_get_traceparent_fields_string_input_no_dash(self):
        traceparent = "invalid"
        version, trace_id, parent_id, trace_flags = self.tp.get_traceparent_fields(traceparent)
        self.assertIsNone(trace_id)
        self.assertIsNone(parent_id)

    def test_update_traceparent(self):
        traceparent = "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"
        in_trace_id = "1234d0e0e4736234"
        in_span_id = "1234567890abcdef"
        level = 1
        expected_traceparent = "00-4bf92f3577b34da6a3ce929d0e0e4736-1234567890abcdef-01"
        self.assertEqual(expected_traceparent, self.tp.update_traceparent(traceparent, in_trace_id, in_span_id, level))

    def test_update_traceparent_None(self):
        traceparent = None
        in_trace_id = "1234d0e0e4736234"
        in_span_id = "7890abcdef"
        level = 0
        expected_traceparent = "00-00000000000000001234d0e0e4736234-0000007890abcdef-00"
        self.assertEqual(expected_traceparent, self.tp.update_traceparent(traceparent, in_trace_id, in_span_id, level))



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

from instana.log import logger
from instana.propagators.base_propagator import BasePropagator


class BinaryPropagator(BasePropagator):
    """
    A Propagator for BINARY.
    The BINARY format represents SpanContexts in an opaque bytearray carrier.
    """

    # ByteArray variations from base class
    HEADER_KEY_T = b'x-instana-t'
    HEADER_KEY_S = b'x-instana-s'
    HEADER_KEY_L = b'x-instana-l'
    HEADER_SERVER_TIMING = b'server-timing'
    HEADER_KEY_TRACEPARENT = b'traceparent'
    HEADER_KEY_TRACESTATE = b'tracestate'

    def __init__(self):
        super(BinaryPropagator, self).__init__()

    def inject(self, span_context, carrier, disable_w3c_trace_context=True):
        try:
            trace_id = str.encode(span_context.trace_id)
            span_id = str.encode(span_context.span_id)
            level = str.encode(str(span_context.level))
            server_timing = str.encode("intid;desc=%s" % span_context.trace_id)

            if disable_w3c_trace_context:
                traceparent, tracestate = [None] * 2
            else:
                traceparent, tracestate = self._get_participating_trace_context(span_context)
                try:
                    traceparent = str.encode(traceparent)
                    tracestate = str.encode(tracestate)
                except Exception:
                    traceparent, tracestate = [None] * 2

            if isinstance(carrier, dict) or hasattr(carrier, "__dict__"):
                if traceparent and tracestate:
                    carrier[self.HEADER_KEY_TRACEPARENT] = traceparent
                    carrier[self.HEADER_KEY_TRACESTATE] = tracestate
                carrier[self.HEADER_KEY_T] = trace_id
                carrier[self.HEADER_KEY_S] = span_id
                carrier[self.HEADER_KEY_L] = level
                carrier[self.HEADER_SERVER_TIMING] = server_timing
            elif isinstance(carrier, list):
                if traceparent and tracestate:
                    carrier.append((self.HEADER_KEY_TRACEPARENT, traceparent))
                    carrier.append((self.HEADER_KEY_TRACESTATE, tracestate))
                carrier.append((self.HEADER_KEY_T, trace_id))
                carrier.append((self.HEADER_KEY_S, span_id))
                carrier.append((self.HEADER_KEY_L, level))
                carrier.append((self.HEADER_SERVER_TIMING, server_timing))
            elif isinstance(carrier, tuple):
                if traceparent and tracestate:
                    carrier = carrier.__add__(((self.HEADER_KEY_TRACEPARENT, traceparent),))
                    carrier = carrier.__add__(((self.HEADER_KEY_TRACESTATE, tracestate),))
                carrier = carrier.__add__(((self.HEADER_KEY_T, trace_id),))
                carrier = carrier.__add__(((self.HEADER_KEY_S, span_id),))
                carrier = carrier.__add__(((self.HEADER_KEY_L, level),))
                carrier = carrier.__add__(((self.HEADER_SERVER_TIMING, server_timing),))
            elif hasattr(carrier, '__setitem__'):
                if traceparent and tracestate:
                    carrier.__setitem__(self.HEADER_KEY_TRACEPARENT, traceparent)
                    carrier.__setitem__(self.HEADER_KEY_TRACESTATE, tracestate)
                carrier.__setitem__(self.HEADER_KEY_T, trace_id)
                carrier.__setitem__(self.HEADER_KEY_S, span_id)
                carrier.__setitem__(self.HEADER_KEY_L, level)
                carrier.__setitem__(self.HEADER_SERVER_TIMING, server_timing)
            else:
                raise Exception("Unsupported carrier type", type(carrier))

            return carrier
        except Exception:
            logger.debug("inject error:", exc_info=True)





# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

from __future__ import absolute_import

import os
import sys

import opentracing as ot
import opentracing.ext.tags as ext
import wrapt

from ...log import logger
from ...singletons import agent, tracer
from ...util.secrets import strip_secrets_from_query

DJ_INSTANA_MIDDLEWARE = 'instana.instrumentation.django.middleware.InstanaMiddleware'

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object


class InstanaMiddleware(MiddlewareMixin):
    """ Django Middleware to provide request tracing for Instana """

    def __init__(self, get_response=None):
        super(InstanaMiddleware, self).__init__(get_response)
        self.get_response = get_response

    def process_request(self, request):
        try:
            env = request.environ

            ctx = tracer.extract(ot.Format.HTTP_HEADERS, env)
            request.iscope = tracer.start_active_span('django', child_of=ctx)

            if agent.options.extra_http_headers is not None:
                for custom_header in agent.options.extra_http_headers:
                    # Headers are available in this format: HTTP_X_CAPTURE_THIS
                    django_header = ('HTTP_' + custom_header.upper()).replace('-', '_')
                    if django_header in env:
                        request.iscope.span.set_tag("http.header.%s" % custom_header, env[django_header])

            request.iscope.span.set_tag(ext.HTTP_METHOD, request.method)
            if 'PATH_INFO' in env:
                request.iscope.span.set_tag(ext.HTTP_URL, env['PATH_INFO'])
            if 'QUERY_STRING' in env and len(env['QUERY_STRING']):
                scrubbed_params = strip_secrets_from_query(env['QUERY_STRING'], agent.options.secrets_matcher,
                                                           agent.options.secrets_list)
                request.iscope.span.set_tag("http.params", scrubbed_params)
            if 'HTTP_HOST' in env:
                request.iscope.span.set_tag("http.host", env['HTTP_HOST'])
        except Exception:
            logger.debug("Django middleware @ process_request", exc_info=True)

    def process_response(self, request, response):
        try:
            if request.iscope is not None:
                if 500 <= response.status_code <= 511:
                    request.iscope.span.assure_errored()
                # for django >= 2.2
                if request.resolver_match is not None and hasattr(request.resolver_match, 'route'):
                    path_tpl = request.resolver_match.route
                # django < 2.2 or in case of 404
                else:
                    try:
                        from django.urls import resolve
                        view_name = resolve(request.path)._func_path
                        path_tpl = "".join(self.__url_pattern_route(view_name))
                    except Exception:
                        # the resolve method can fire a Resolver404 exception, in this case there is no matching route
                        # so the path_tpl is set to None in order not to be added as a tag
                        path_tpl = None
                if path_tpl:
                    request.iscope.span.set_tag("http.path_tpl", path_tpl)
                request.iscope.span.set_tag(ext.HTTP_STATUS_CODE, response.status_code)
                tracer.inject(request.iscope.span.context, ot.Format.HTTP_HEADERS, response)
                response['Server-Timing'] = "intid;desc=%s" % request.iscope.span.context.trace_id
        except Exception:
            logger.debug("Instana middleware @ process_response", exc_info=True)
        finally:
            if request.iscope is not None:
                request.iscope.close()
                request.iscope = None
        return response

    def process_exception(self, request, exception):
        from django.http.response import Http404

        if isinstance(exception, Http404):
            return None

        if request.iscope is not None:
            request.iscope.span.log_exception(exception)

    def __url_pattern_route(self, view_name):
        from django.conf import settings
        try:
            from django.urls import RegexURLPattern as URLPattern
            from django.urls import RegexURLResolver as URLResolver
        except ImportError:
            from django.urls import URLPattern, URLResolver

        urlconf = __import__(settings.ROOT_URLCONF, {}, {}, [''])

        def list_urls(urlpatterns, parent_pattern=None):
            if not urlpatterns:
                return
            if parent_pattern is None:
                parent_pattern = []
            first = urlpatterns[0]
            if isinstance(first, URLPattern):
                if first.lookup_str == view_name:
                    if hasattr(first, "regex"):
                        return parent_pattern + [str(first.regex.pattern)]
                    else:
                        return parent_pattern + [str(first.pattern)]
            elif isinstance(first, URLResolver):
                if hasattr(first, "regex"):
                    return list_urls(first.url_patterns, parent_pattern + [str(first.regex.pattern)])
                else:
                    return list_urls(first.url_patterns, parent_pattern + [str(first.pattern)])
            return list_urls(urlpatterns[1:], parent_pattern)

        return list_urls(urlconf.urlpatterns)


def load_middleware_wrapper(wrapped, instance, args, kwargs):
    try:
        from django.conf import settings

        # Django >=1.10 to <2.0 support old-style MIDDLEWARE_CLASSES so we
        # do as well here
        if hasattr(settings, 'MIDDLEWARE') and settings.MIDDLEWARE is not None:
            if DJ_INSTANA_MIDDLEWARE in settings.MIDDLEWARE:
                return wrapped(*args, **kwargs)

            if isinstance(settings.MIDDLEWARE, tuple):
                settings.MIDDLEWARE = (DJ_INSTANA_MIDDLEWARE,) + settings.MIDDLEWARE
            elif isinstance(settings.MIDDLEWARE, list):
                settings.MIDDLEWARE = [DJ_INSTANA_MIDDLEWARE] + settings.MIDDLEWARE
            else:
                logger.warning("Instana: Couldn't add InstanaMiddleware to Django")

        elif hasattr(settings, 'MIDDLEWARE_CLASSES') and settings.MIDDLEWARE_CLASSES is not None:
            if DJ_INSTANA_MIDDLEWARE in settings.MIDDLEWARE_CLASSES:
                return wrapped(*args, **kwargs)

            if isinstance(settings.MIDDLEWARE_CLASSES, tuple):
                settings.MIDDLEWARE_CLASSES = (DJ_INSTANA_MIDDLEWARE,) + settings.MIDDLEWARE_CLASSES
            elif isinstance(settings.MIDDLEWARE_CLASSES, list):
                settings.MIDDLEWARE_CLASSES = [DJ_INSTANA_MIDDLEWARE] + settings.MIDDLEWARE_CLASSES
            else:
                logger.warning("Instana: Couldn't add InstanaMiddleware to Django")

        else:
            logger.warning("Instana: Couldn't find middleware settings")

        return wrapped(*args, **kwargs)
    except Exception:
        logger.warning("Instana: Couldn't add InstanaMiddleware to Django: ", exc_info=True)


try:
    if 'django' in sys.modules:
        logger.debug("Instrumenting django")
        wrapt.wrap_function_wrapper('django.core.handlers.base', 'BaseHandler.load_middleware', load_middleware_wrapper)

        if 'INSTANA_MAGIC' in os.environ:
            # If we are instrumenting via AutoTrace (in an already running process), then the
            # WSGI middleware has to be live reloaded.
            from django.core.servers.basehttp import get_internal_wsgi_application
            from django.core.exceptions import ImproperlyConfigured

            try:
                wsgiapp = get_internal_wsgi_application()
                wsgiapp.load_middleware()
            except ImproperlyConfigured:
                pass

except Exception:
    logger.debug("django.middleware:", exc_info=True)
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021


from sanic.views import HTTPMethodView
from sanic.response import text

class SimpleView(HTTPMethodView):

  def get(self, request):
      return text("I am get method")

  # You can also use async syntax
  async def post(self, request):
      return text("I am post method")

  def put(self, request):
      return text("I am put method")

  def patch(self, request):
      return text("I am patch method")

  def delete(self, request):
      return text("I am delete method")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

"""
The Instana agent (for GCR) that manages
monitoring state and reporting that data.
"""
import time
from instana.options import GCROptions
from instana.collector.google_cloud_run import GCRCollector
from instana.log import logger
from instana.util import to_json
from instana.agent.base import BaseAgent
from instana.version import VERSION


class GCRAgent(BaseAgent):
    """ In-process agent for Google Cloud Run """

    def __init__(self, service, configuration, revision):
        super(GCRAgent, self).__init__()

        self.options = GCROptions()
        self.collector = None
        self.report_headers = None
        self._can_send = False

        # Update log level (if INSTANA_LOG_LEVEL was set)
        self.update_log_level()

        logger.info("Stan is on the AWS Fargate scene.  Starting Instana instrumentation version: %s", VERSION)

        if self._validate_options():
            self._can_send = True
            self.collector = GCRCollector(self, service, configuration, revision)
            self.collector.start()
        else:
            logger.warning("Required INSTANA_AGENT_KEY and/or INSTANA_ENDPOINT_URL environment variables not set.  "
                           "We will not be able monitor this GCR cluster.")

    def can_send(self):
        """
        Are we in a state where we can send data?
        @return: Boolean
        """
        return self._can_send

    def get_from_structure(self):
        """
        Retrieves the From data that is reported alongside monitoring data.
        @return: dict()
        """
        return {'hl': True, 'cp': 'gcp', 'e': self.collector.get_instance_id()}

    def report_data_payload(self, payload):
        """
        Used to report metrics and span data to the endpoint URL in self.options.endpoint_url
        """
        response = None
        try:
            if self.report_headers is None:
                # Prepare request headers
                self.report_headers = {
                    "Content-Type": "application/json",
                    "X-Instana-Host": "gcp:cloud-run:revision:{revision}".format(
                        revision=self.collector.revision),
                    "X-Instana-Key": self.options.agent_key
                }

            self.report_headers["X-Instana-Time"] = str(round(time.time() * 1000))

            response = self.client.post(self.__data_bundle_url(),
                                        data=to_json(payload),
                                        headers=self.report_headers,
                                        timeout=self.options.timeout,
                                        verify=self.options.ssl_verify,
                                        proxies=self.options.endpoint_proxy)

            if response.status_code >= 400:
                logger.info("report_data_payload: Instana responded with status code %s", response.status_code)
        except Exception as exc:
            logger.debug("report_data_payload: connection error (%s)", type(exc))
        return response

    def _validate_options(self):
        """
        Validate that the options used by this Agent are valid.  e.g. can we report data?
        """
        return self.options.endpoint_url is not None and self.options.agent_key is not None

    def __data_bundle_url(self):
        """
        URL for posting metrics to the host agent.  Only valid when announced.
        """
        return "{endpoint_url}/bundle".format(endpoint_url=self.options.endpoint_url)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import instana

from sanic import Sanic
from sanic.exceptions import SanicException
from tests.apps.sanic_app.simpleview import SimpleView
from tests.apps.sanic_app.name import NameView
from sanic.response import text

app = Sanic('test')

@app.get("/foo/<foo_id:int>")
async def uuid_handler(request, foo_id: int):
    return text("INT - {}".format(foo_id))


@app.route("/test_request_args")
async def test_request_args(request):
    raise SanicException("Something went wrong.", status_code=500)

@app.route("/instana_exception")
async def test_request_args(request):
    raise SanicException(description="Something went wrong.", status_code=500)

@app.route("/wrong")
async def test_request_args(request):
    raise SanicException(message="Something went wrong.", status_code=400)

@app.get("/tag/<tag>")
async def tag_handler(request, tag):
    return text("Tag - {}".format(tag))


app.add_route(SimpleView.as_view(), "/")
app.add_route(NameView.as_view(), "/<name>")


if __name__ == '__main__':
    app.run(host="0.0.0.0", port=8000, debug=True, access_log=True)






# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from nose.plugins.skip import SkipTest
from opentracing.harness.api_check import APICompatibilityCheckMixin

from instana.tracer import InstanaTracer


class TestInstanaTracer(InstanaTracer, APICompatibilityCheckMixin):
    def tracer(self):
        return self

    def test_binary_propagation(self):
        raise SkipTest('Binary format is not supported')

    def test_mandatory_formats(self):
        raise SkipTest('Binary format is not supported')

    def check_baggage_values(self):
        return True

    def is_parent(self, parent, span):
        # use `Span` ids to check parenting
        if parent is None:
            return span.parent_id is None

        return parent.context.span_id == span.parent_id



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instana WSGI Middleware
"""
import opentracing as ot
import opentracing.ext.tags as tags

from ..singletons import agent, tracer
from ..util.secrets import strip_secrets_from_query


class InstanaWSGIMiddleware(object):
    """ Instana WSGI middleware """

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        env = environ

        def new_start_response(status, headers, exc_info=None):
            """Modified start response with additional headers."""
            tracer.inject(self.scope.span.context, ot.Format.HTTP_HEADERS, headers)
            headers.append(('Server-Timing', "intid;desc=%s" % self.scope.span.context.trace_id))

            res = start_response(status, headers, exc_info)

            sc = status.split(' ')[0]
            if 500 <= int(sc) <= 511:
                self.scope.span.mark_as_errored()

            self.scope.span.set_tag(tags.HTTP_STATUS_CODE, sc)
            self.scope.close()
            return res

        ctx = tracer.extract(ot.Format.HTTP_HEADERS, env)
        self.scope = tracer.start_active_span("wsgi", child_of=ctx)

        if agent.options.extra_http_headers is not None:
            for custom_header in agent.options.extra_http_headers:
                # Headers are available in this format: HTTP_X_CAPTURE_THIS
                wsgi_header = ('HTTP_' + custom_header.upper()).replace('-', '_')
                if wsgi_header in env:
                    self.scope.span.set_tag("http.header.%s" % custom_header, env[wsgi_header])

        if 'PATH_INFO' in env:
            self.scope.span.set_tag('http.path', env['PATH_INFO'])
        if 'QUERY_STRING' in env and len(env['QUERY_STRING']):
            scrubbed_params = strip_secrets_from_query(env['QUERY_STRING'], agent.options.secrets_matcher,
                                                       agent.options.secrets_list)
            self.scope.span.set_tag("http.params", scrubbed_params)
        if 'REQUEST_METHOD' in env:
            self.scope.span.set_tag(tags.HTTP_METHOD, env['REQUEST_METHOD'])
        if 'HTTP_HOST' in env:
            self.scope.span.set_tag("http.host", env['HTTP_HOST'])

        return self.app(environ, new_start_response)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

""" Module to handle the collection of container metrics in AWS Fargate """
from ....log import logger
from ....util import DictionaryOfStan
from ..base import BaseHelper


class ContainerHelper(BaseHelper):
    """ This class acts as a helper to collect container snapshot and metric information """
    def collect_metrics(self, **kwargs):
        """
        Collect and return metrics (and optionally snapshot data) for every container in this task
        @return: list - with one or more plugin entities
        """
        plugins = []

        try:
            if self.collector.task_metadata is not None:
                containers = self.collector.task_metadata.get("Containers", [])
                for container in containers:
                    plugin_data = dict()
                    plugin_data["name"] = "com.instana.plugin.aws.ecs.container"
                    try:
                        labels = container.get("Labels", {})
                        name = container.get("Name", "")
                        task_arn = labels.get("com.amazonaws.ecs.task-arn", "")
                        plugin_data["entityId"] = "%s::%s" % (task_arn, name)

                        plugin_data["data"] = DictionaryOfStan()
                        if self.collector.root_metadata["Name"] == name:
                            plugin_data["data"]["instrumented"] = True
                        plugin_data["data"]["dockerId"] = container.get("DockerId", None)
                        plugin_data["data"]["taskArn"] = labels.get("com.amazonaws.ecs.task-arn", None)

                        if kwargs.get("with_snapshot"):
                            plugin_data["data"]["runtime"] = "python"
                            plugin_data["data"]["dockerName"] = container.get("DockerName", None)
                            plugin_data["data"]["containerName"] = container.get("Name", None)
                            plugin_data["data"]["image"] = container.get("Image", None)
                            plugin_data["data"]["imageId"] = container.get("ImageID", None)
                            plugin_data["data"]["taskDefinition"] = labels.get("com.amazonaws.ecs.task-definition-family", None)
                            plugin_data["data"]["taskDefinitionVersion"] = labels.get("com.amazonaws.ecs.task-definition-version", None)
                            plugin_data["data"]["clusterArn"] = labels.get("com.amazonaws.ecs.cluster", None)
                            plugin_data["data"]["desiredStatus"] = container.get("DesiredStatus", None)
                            plugin_data["data"]["knownStatus"] = container.get("KnownStatus", None)
                            plugin_data["data"]["ports"] = container.get("Ports", None)
                            plugin_data["data"]["createdAt"] = container.get("CreatedAt", None)
                            plugin_data["data"]["startedAt"] = container.get("StartedAt", None)
                            plugin_data["data"]["type"] = container.get("Type", None)
                            limits = container.get("Limits", {})
                            plugin_data["data"]["limits"]["cpu"] = limits.get("CPU", None)
                            plugin_data["data"]["limits"]["memory"] = limits.get("Memory", None)
                    except Exception:
                        logger.debug("_collect_container_snapshots: ", exc_info=True)
                    finally:
                        plugins.append(plugin_data)
        except Exception:
            logger.debug("collect_container_metrics: ", exc_info=True)
        return plugins



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from ..log import logger


class InstanaAncestor:
    def __init__(self, trace_id, parent_id):
        self.t = trace_id
        self.p = parent_id


class Tracestate:
    MAX_NUMBER_OF_LIST_MEMBERS = 32
    REMOVE_ENTRIES_LARGER_THAN = 128

    @staticmethod
    def get_instana_ancestor(tracestate):
        """
        Constructs the instana ancestor object and returns it
        :param tracestate: the original tracestate value
        :return: instana ancestor instance
        """
        try:
            in_list_member = tracestate.strip().split("in=")[1].split(",")[0]

            ia = InstanaAncestor(trace_id=in_list_member.split(";")[0],
                                 parent_id=in_list_member.split(";")[1])
            return ia

        except Exception:
            logger.debug("extract instana ancestor error:", exc_info=True)
        return None

    def update_tracestate(self, tracestate, in_trace_id, in_span_id):
        """
        Method to update the tracestate property with the instana trace_id and span_id

        :param tracestate: original tracestate header
        :param in_trace_id: instana trace_id
        :param in_span_id: instana parent_id
        :return: tracestate updated
        """
        try:
            span_id = in_span_id.zfill(16)  # if span_id is shorter than 16 characters we prepend zeros
            instana_tracestate = "in={};{}".format(in_trace_id, span_id)
            if tracestate is None or tracestate == "":
                tracestate = instana_tracestate
            else:
                # remove the existing in= entry
                if "in=" in tracestate:
                    splitted = tracestate.split("in=")
                    before_in = splitted[0]
                    after_in = splitted[1].split(",")[1:]
                    tracestate = '{}{}'.format(before_in, ",".join(after_in))
                # tracestate can contain a max of 32 list members, if it contains up to 31
                # we can safely add the instana one without the need to truncate anything
                if len(tracestate.split(",")) <= self.MAX_NUMBER_OF_LIST_MEMBERS - 1:
                    tracestate = "{},{}".format(instana_tracestate, tracestate)
                else:
                    list_members = tracestate.split(",")
                    list_members_to_remove = len(list_members) - self.MAX_NUMBER_OF_LIST_MEMBERS + 1
                    # Number 1 priority members to be removed are the ones larger than 128 characters
                    for i, m in reversed(list(enumerate(list_members))):
                        if len(m) > self.REMOVE_ENTRIES_LARGER_THAN:
                            list_members.pop(i)
                            list_members_to_remove -= 1
                        if list_members_to_remove == 0:
                            break
                    # if there are still more than 31 list members remaining, we remove as many members
                    # from the end as necessary to remain just 31 list members
                    while list_members_to_remove > 0:
                        list_members.pop()
                        list_members_to_remove -= 1
                    # update the tracestate containing just 31 list members
                    tracestate = ",".join(list_members)
                    # adding instana as first list member, total of 32 list members
                    tracestate = "{},{}".format(instana_tracestate, tracestate)
        except Exception:
            logger.debug("Something went wrong while updating tracestate: {}:".format(tracestate), exc_info=True)

        return tracestate



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

""" Collection helper for the process """
import os
import pwd
import grp
from instana.log import logger
from instana.util import DictionaryOfStan
from instana.util.runtime import get_proc_cmdline
from instana.util.secrets import contains_secret
from .base import BaseHelper


class ProcessHelper(BaseHelper):
    """ Helper class to collect metrics for this process """

    def collect_metrics(self, **kwargs):
        plugin_data = dict()
        try:
            plugin_data["name"] = "com.instana.plugin.process"
            plugin_data["entityId"] = str(os.getpid())
            plugin_data["data"] = DictionaryOfStan()
            plugin_data["data"]["pid"] = int(os.getpid())

            if kwargs.get("with_snapshot"):
                self._collect_process_snapshot(plugin_data)
        except Exception:
            logger.debug("ProcessHelper.collect_metrics: ", exc_info=True)
        return plugin_data

    def _collect_process_snapshot(self, plugin_data):
        try:
            env = dict()
            for key in os.environ:
                if contains_secret(key,
                                   self.collector.agent.options.secrets_matcher,
                                   self.collector.agent.options.secrets_list):
                    env[key] = "<ignored>"
                else:
                    env[key] = os.environ[key]
            plugin_data["data"]["env"] = env
            if os.path.isfile("/proc/self/exe"):
                plugin_data["data"]["exec"] = os.readlink("/proc/self/exe")
            else:
                logger.debug("Can't access /proc/self/exe...")

            cmdline = get_proc_cmdline()
            if len(cmdline) > 1:
                # drop the exe
                cmdline.pop(0)
            plugin_data["data"]["args"] = cmdline
            try:
                euid = os.geteuid()
                egid = os.getegid()
                plugin_data["data"]["user"] = pwd.getpwuid(euid)
                plugin_data["data"]["group"] = grp.getgrgid(egid).gr_name
            except Exception:
                logger.debug("euid/egid detection: ", exc_info=True)

            plugin_data["data"]["start"] = self.collector.fetching_start_time

        except Exception:
            logger.debug("ProcessHelper._collect_process_snapshot: ", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import urllib3
from django.apps import apps
from ..apps.app_django import INSTALLED_APPS
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
import os
from instana.singletons import agent, tracer

from ..helpers import fail_with_message_and_span_dump, get_first_span_by_filter, drop_log_spans_from_list

apps.populate(INSTALLED_APPS)


class TestDjango(StaticLiveServerTestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = tracer.recorder
        self.recorder.clear_spans()
        self.http = urllib3.PoolManager()

    def tearDown(self):
        """ Do nothing for now """
        os.environ["INSTANA_DISABLE_W3C_TRACE_CORRELATION"] = ""

    def test_basic_request(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', self.live_server_url + '/', fields={"test": 1})

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        test_span = spans[2]
        urllib3_span = spans[1]
        django_span = spans[0]

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual("django", django_span.n)

        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, django_span.t)

        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(django_span.p, urllib3_span.s)

        self.assertIsNone(django_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

        self.assertEqual(None, django_span.ec)
        self.assertEqual('/', django_span.data["http"]["url"])
        self.assertEqual('GET', django_span.data["http"]["method"])
        self.assertEqual(200, django_span.data["http"]["status"])
        self.assertEqual('test=1', django_span.data["http"]["params"])
        self.assertEqual('^$', django_span.data["http"]["path_tpl"])

        self.assertIsNone(django_span.stack)

    def test_synthetic_request(self):
        headers = {
            'X-INSTANA-SYNTHETIC': '1'
        }

        with tracer.start_active_span('test'):
            response = self.http.request('GET', self.live_server_url + '/', headers=headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        test_span = spans[2]
        urllib3_span = spans[1]
        django_span = spans[0]

        self.assertEqual('^$', django_span.data["http"]["path_tpl"])

        self.assertTrue(django_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

    def test_request_with_error(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', self.live_server_url + '/cause_error')

        assert response
        self.assertEqual(500, response.status)

        spans = self.recorder.queued_spans()
        spans = drop_log_spans_from_list(spans)

        span_count = len(spans)
        if span_count != 3:
            msg = "Expected 3 spans but got %d" % span_count
            fail_with_message_and_span_dump(msg, spans)

        filter = lambda span: span.n == 'sdk' and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, filter)
        assert (test_span)

        filter = lambda span: span.n == 'urllib3'
        urllib3_span = get_first_span_by_filter(spans, filter)
        assert (urllib3_span)

        filter = lambda span: span.n == 'django'
        django_span = get_first_span_by_filter(spans, filter)
        assert (django_span)

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual("django", django_span.n)

        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, django_span.t)

        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(django_span.p, urllib3_span.s)

        self.assertEqual(1, django_span.ec)

        self.assertEqual('/cause_error', django_span.data["http"]["url"])
        self.assertEqual('GET', django_span.data["http"]["method"])
        self.assertEqual(500, django_span.data["http"]["status"])
        self.assertEqual('This is a fake error: /cause-error', django_span.data["http"]["error"])
        self.assertEqual('^cause_error$', django_span.data["http"]["path_tpl"])
        self.assertIsNone(django_span.stack)

    def test_request_with_not_found(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', self.live_server_url + '/not_found')

        assert response
        self.assertEqual(404, response.status)

        spans = self.recorder.queued_spans()
        spans = drop_log_spans_from_list(spans)

        span_count = len(spans)
        if span_count != 3:
            msg = "Expected 3 spans but got %d" % span_count
            fail_with_message_and_span_dump(msg, spans)

        filter = lambda span: span.n == 'django'
        django_span = get_first_span_by_filter(spans, filter)
        assert (django_span)

        self.assertIsNone(django_span.ec)
        self.assertEqual(404, django_span.data["http"]["status"])

    def test_request_with_not_found_no_route(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', self.live_server_url + '/no_route')

        assert response
        self.assertEqual(404, response.status)

        spans = self.recorder.queued_spans()
        spans = drop_log_spans_from_list(spans)

        span_count = len(spans)
        if span_count != 3:
            msg = "Expected 3 spans but got %d" % span_count
            fail_with_message_and_span_dump(msg, spans)

        filter = lambda span: span.n == 'django'
        django_span = get_first_span_by_filter(spans, filter)
        assert (django_span)
        self.assertIsNone(django_span.data["http"]["path_tpl"])
        self.assertIsNone(django_span.ec)
        self.assertEqual(404, django_span.data["http"]["status"])

    def test_complex_request(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', self.live_server_url + '/complex')

        assert response
        self.assertEqual(200, response.status)
        spans = self.recorder.queued_spans()
        self.assertEqual(5, len(spans))

        test_span = spans[4]
        urllib3_span = spans[3]
        django_span = spans[2]
        ot_span1 = spans[1]
        ot_span2 = spans[0]

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual("django", django_span.n)
        self.assertEqual("sdk", ot_span1.n)
        self.assertEqual("sdk", ot_span2.n)

        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, django_span.t)
        self.assertEqual(django_span.t, ot_span1.t)
        self.assertEqual(ot_span1.t, ot_span2.t)

        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(django_span.p, urllib3_span.s)
        self.assertEqual(ot_span1.p, django_span.s)
        self.assertEqual(ot_span2.p, ot_span1.s)

        self.assertEqual(None, django_span.ec)
        self.assertIsNone(django_span.stack)

        self.assertEqual('/complex', django_span.data["http"]["url"])
        self.assertEqual('GET', django_span.data["http"]["method"])
        self.assertEqual(200, django_span.data["http"]["status"])
        self.assertEqual('^complex$', django_span.data["http"]["path_tpl"])

    def test_custom_header_capture(self):
        # Hack together a manual custom headers list
        agent.options.extra_http_headers = [u'X-Capture-This', u'X-Capture-That']

        request_headers = dict()
        request_headers['X-Capture-This'] = 'this'
        request_headers['X-Capture-That'] = 'that'

        with tracer.start_active_span('test'):
            response = self.http.request('GET', self.live_server_url + '/', headers=request_headers)
            # response = self.client.get('/')

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        test_span = spans[2]
        urllib3_span = spans[1]
        django_span = spans[0]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual("django", django_span.n)

        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, django_span.t)

        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(django_span.p, urllib3_span.s)

        self.assertEqual(None, django_span.ec)
        self.assertIsNone(django_span.stack)

        self.assertEqual('/', django_span.data["http"]["url"])
        self.assertEqual('GET', django_span.data["http"]["method"])
        self.assertEqual(200, django_span.data["http"]["status"])
        self.assertEqual('^$', django_span.data["http"]["path_tpl"])

        assert "X-Capture-This" in django_span.data["http"]["header"]
        self.assertEqual("this", django_span.data["http"]["header"]["X-Capture-This"])
        assert "X-Capture-That" in django_span.data["http"]["header"]
        self.assertEqual("that", django_span.data["http"]["header"]["X-Capture-That"])

    def test_with_incoming_context(self):
        request_headers = dict()
        request_headers['X-INSTANA-T'] = '1'
        request_headers['X-INSTANA-S'] = '1'
        request_headers['traceparent'] = '01-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01-788777'
        request_headers['tracestate'] = 'rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE'

        response = self.http.request('GET', self.live_server_url + '/', headers=request_headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        django_span = spans[0]

        self.assertEqual(django_span.t, '0000000000000001')
        self.assertEqual(django_span.p, '0000000000000001')

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        assert ('traceparent' in response.headers)
        self.assertEqual('01-4bf92f3577b34da6a3ce929d0e0e4736-{}-01'.format(django_span.s),
                         response.headers['traceparent'])

        assert ('tracestate' in response.headers)
        self.assertEqual(
            'in={};{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE'.format(
                django_span.t, django_span.s), response.headers['tracestate'])
        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])

    def test_with_incoming_context_and_correlation(self):
        request_headers = dict()
        request_headers['X-INSTANA-T'] = '1'
        request_headers['X-INSTANA-S'] = '1'
        request_headers['X-INSTANA-L'] = '1, correlationType=web; correlationId=1234567890abcdef'
        request_headers['traceparent'] = '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'
        request_headers['tracestate'] = 'rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE'

        response = self.http.request('GET', self.live_server_url + '/', headers=request_headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        django_span = spans[0]

        self.assertEqual(django_span.t, 'a3ce929d0e0e4736')
        self.assertEqual(django_span.p, '00f067aa0ba902b7')
        self.assertEqual(django_span.ia.t, 'a3ce929d0e0e4736')
        self.assertEqual(django_span.ia.p, '8357ccd9da194656')
        self.assertEqual(django_span.lt, '4bf92f3577b34da6a3ce929d0e0e4736')
        self.assertEqual(django_span.tp, True)
        self.assertEqual(django_span.crtp, 'web')
        self.assertEqual(django_span.crid, '1234567890abcdef')

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        assert ('traceparent' in response.headers)
        self.assertEqual('00-4bf92f3577b34da6a3ce929d0e0e4736-{}-01'.format(django_span.s),
                         response.headers['traceparent'])

        assert ('tracestate' in response.headers)
        self.assertEqual(
            'in={};{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE'.format(
                django_span.t, django_span.s), response.headers['tracestate'])
        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])

    def test_with_incoming_traceparent_tracestate(self):
        request_headers = dict()
        request_headers['traceparent'] = '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'
        request_headers['tracestate'] = 'rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE'

        response = self.http.request('GET', self.live_server_url + '/', headers=request_headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        django_span = spans[0]

        self.assertEqual(django_span.t, 'a3ce929d0e0e4736')  # last 16 chars from traceparent trace_id
        self.assertEqual(django_span.p, '00f067aa0ba902b7')
        self.assertEqual(django_span.ia.t, 'a3ce929d0e0e4736')
        self.assertEqual(django_span.ia.p, '8357ccd9da194656')
        self.assertEqual(django_span.lt, '4bf92f3577b34da6a3ce929d0e0e4736')
        self.assertEqual(django_span.tp, True)

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        assert ('traceparent' in response.headers)
        self.assertEqual('00-4bf92f3577b34da6a3ce929d0e0e4736-{}-01'.format(django_span.s),
                         response.headers['traceparent'])

        assert ('tracestate' in response.headers)
        self.assertEqual(
            'in=a3ce929d0e0e4736;{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE'.format(
                django_span.s), response.headers['tracestate'])

        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])

    def test_with_incoming_traceparent_tracestate_disable_traceparent(self):
        os.environ["INSTANA_DISABLE_W3C_TRACE_CORRELATION"] = "1"
        request_headers = dict()
        request_headers['traceparent'] = '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'
        request_headers['tracestate'] = 'rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE'

        response = self.http.request('GET', self.live_server_url + '/', headers=request_headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        django_span = spans[0]

        self.assertEqual(django_span.t, 'a3ce929d0e0e4736')  # last 16 chars from traceparent trace_id
        self.assertEqual(django_span.p, '8357ccd9da194656')

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        assert ('traceparent' in response.headers)
        self.assertEqual('00-4bf92f3577b34da6a3ce929d0e0e4736-{}-01'.format(django_span.s),
                         response.headers['traceparent'])

        assert ('tracestate' in response.headers)
        self.assertEqual(
            'in=a3ce929d0e0e4736;{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE'.format(
                django_span.s), response.headers['tracestate'])

        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])

    def test_with_incoming_mixed_case_context(self):
        request_headers = dict()
        request_headers['X-InSTANa-T'] = '0000000000000001'
        request_headers['X-instana-S'] = '0000000000000001'

        response = self.http.request('GET', self.live_server_url + '/', headers=request_headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        django_span = spans[0]

        self.assertEqual(django_span.t, '0000000000000001')
        self.assertEqual(django_span.p, '0000000000000001')

        assert ('X-INSTANA-T' in response.headers)
        assert (int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(django_span.t, response.headers['X-INSTANA-T'])

        assert ('X-INSTANA-S' in response.headers)
        assert (int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(django_span.s, response.headers['X-INSTANA-S'])

        assert ('X-INSTANA-L' in response.headers)
        self.assertEqual('1', response.headers['X-INSTANA-L'])

        server_timing_value = "intid;desc=%s" % django_span.t
        assert ('Server-Timing' in response.headers)
        self.assertEqual(server_timing_value, response.headers['Server-Timing'])



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import os
import sys
import pytest
import asynqp
import asyncio
import aiohttp
import unittest
import opentracing
from distutils.version import LooseVersion

import tests.apps.flask_app
from ..helpers import testenv
from instana.singletons import async_tracer

rabbitmq_host = ""
if "RABBITMQ_HOST" in os.environ:
    rabbitmq_host = os.environ["RABBITMQ_HOST"]
else:
    rabbitmq_host = "localhost"

is_unsupported_version = LooseVersion(sys.version) < LooseVersion('3.5.3') \
                         or LooseVersion(sys.version) >= LooseVersion('3.8.0')


@pytest.mark.skipif(is_unsupported_version, reason="Asynqp supports >=3.5.3;<3.8.0")
class TestAsynqp(unittest.TestCase):
    @asyncio.coroutine
    def connect(self):
        # connect to the RabbitMQ broker
        self.connection = yield from asynqp.connect(rabbitmq_host, 5672, username='guest', password='guest')

        # Open a communications channel
        self.channel = yield from self.connection.open_channel()

        # Create a queue and an exchange on the broker
        self.exchange = yield from self.channel.declare_exchange('test.exchange', 'direct')
        self.queue = yield from self.channel.declare_queue('test.queue')

        # Bind the queue to the exchange, so the queue will get messages published to the exchange
        yield from self.queue.bind(self.exchange, 'routing.key')
        yield from self.queue.purge()

    @asyncio.coroutine
    def reset(self):
        yield from self.queue.delete(if_unused=False, if_empty=False)

    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = async_tracer.recorder
        self.recorder.clear_spans()

        # New event loop for every test
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.loop.run_until_complete(self.connect())

    def tearDown(self):
        """ Purge the queue """
        self.loop.run_until_complete(self.reset())
        self.loop.close()
        self.recorder = async_tracer.recorder
        self.recorder.clear_spans()

    async def fetch(self, session, url, headers=None):
        try:
            async with session.get(url, headers=headers) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    def test_publish(self):
        @asyncio.coroutine
        def test():
            with async_tracer.start_active_span('test'):
                msg = asynqp.Message({'hello': 'world'}, content_type='application/json')
                self.exchange.publish(msg, 'routing.key')

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        rabbitmq_span = spans[0]
        test_span = spans[1]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, rabbitmq_span.t)

        # Parent relationships
        self.assertEqual(rabbitmq_span.p, test_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(rabbitmq_span.ec)

        # Span type
        self.assertEqual(rabbitmq_span.k, 2)  # exit

        # Rabbitmq
        self.assertEqual('test.exchange', rabbitmq_span.data["rabbitmq"]["exchange"])
        self.assertEqual('publish', rabbitmq_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(rabbitmq_span.data["rabbitmq"]["address"])
        self.assertEqual('routing.key', rabbitmq_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(rabbitmq_span.stack)
        self.assertTrue(type(rabbitmq_span.stack) is list)
        self.assertGreater(len(rabbitmq_span.stack), 0)

    def test_many_publishes(self):
        @asyncio.coroutine
        def test():
            @asyncio.coroutine
            def publish_a_bunch(msg):
                for _ in range(20):
                    self.exchange.publish(msg, 'routing.key')

            with async_tracer.start_active_span('test'):
                msg = asynqp.Message({'hello': 'world'})
                yield from publish_a_bunch(msg)

                for _ in range(10):
                    msg = yield from self.queue.get()
                    self.assertIsNotNone(msg)

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertGreaterEqual(len(spans), 31)

        trace_id = spans[0].t
        for span in spans:
            self.assertEqual(span.t, trace_id)

        self.assertIsNone(async_tracer.active_span)

    def test_get(self):
        @asyncio.coroutine
        def publish():
            with async_tracer.start_active_span('test'):
                msg1 = asynqp.Message({'consume': 'this'})
                self.exchange.publish(msg1, 'routing.key')
                msg = yield from self.queue.get()
                self.assertIsNotNone(msg)

        self.loop.run_until_complete(publish())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        publish_span = spans[0]
        get_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, publish_span.t)
        self.assertEqual(test_span.t, get_span.t)

        # Parent relationships
        self.assertEqual(publish_span.p, test_span.s)
        self.assertEqual(get_span.p, test_span.s)

        # Span type
        self.assertEqual(publish_span.k, 2)  # exit
        self.assertEqual(get_span.k, 1)  # entry

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(publish_span.ec)
        self.assertIsNone(get_span.ec)

        # Publish
        self.assertEqual('publish', publish_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(publish_span.data["rabbitmq"]["address"])
        self.assertIsNotNone(publish_span.stack)
        self.assertTrue(type(publish_span.stack) is list)
        self.assertGreater(len(publish_span.stack), 0)

        # get
        self.assertEqual('test.queue', get_span.data["rabbitmq"]["queue"])
        self.assertEqual('consume', get_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(get_span.data["rabbitmq"]["address"])
        self.assertIsNotNone(get_span.stack)
        self.assertTrue(type(get_span.stack) is list)
        self.assertGreater(len(get_span.stack), 0)

    def test_consume(self):
        def handle_message(msg):
            # print('>> {}'.format(msg.body))
            msg.ack()

        @asyncio.coroutine
        def test():
            with async_tracer.start_active_span('test'):
                msg1 = asynqp.Message({'consume': 'this'})
                self.exchange.publish(msg1, 'routing.key')

            self.consumer = yield from self.queue.consume(handle_message)
            yield from asyncio.sleep(0.5)
            self.consumer.cancel()

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        publish_span = spans[0]
        test_span = spans[1]
        consume_span = spans[2]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, publish_span.t)
        self.assertEqual(test_span.t, consume_span.t)

        # Parent relationships
        self.assertEqual(publish_span.p, test_span.s)
        self.assertEqual(consume_span.p, publish_span.s)

        # Span type
        self.assertEqual(publish_span.k, 2)  # exit
        self.assertEqual(consume_span.k, 1)  # entry

        # publish
        self.assertEqual('test.exchange', publish_span.data["rabbitmq"]["exchange"])
        self.assertEqual('publish', publish_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(publish_span.data["rabbitmq"]["address"])
        self.assertEqual('routing.key', publish_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(publish_span.stack)
        self.assertTrue(type(publish_span.stack) is list)
        self.assertGreater(len(publish_span.stack), 0)

        # consume
        self.assertEqual('test.exchange', consume_span.data["rabbitmq"]["exchange"])
        self.assertEqual('consume', consume_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(consume_span.data["rabbitmq"]["address"])
        self.assertEqual('routing.key', consume_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(consume_span.stack)
        self.assertTrue(type(consume_span.stack) is list)
        self.assertGreater(len(consume_span.stack), 0)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(consume_span.ec)
        self.assertIsNone(publish_span.ec)

    def test_consume_and_publish(self):
        def handle_message(msg):
            self.assertIsNotNone(msg)
            msg.ack()
            msg2 = asynqp.Message({'handled': 'msg1'})
            self.exchange.publish(msg2, 'another.key')

        @asyncio.coroutine
        def test():
            with async_tracer.start_active_span('test'):
                msg1 = asynqp.Message({'consume': 'this'})
                self.exchange.publish(msg1, 'routing.key')

                self.consumer = yield from self.queue.consume(handle_message)
                yield from asyncio.sleep(0.5)

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(4, len(spans))

        publish1_span = spans[0]
        publish2_span = spans[1]
        consume1_span = spans[2]
        test_span = spans[3]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, publish1_span.t)
        self.assertEqual(test_span.t, publish2_span.t)
        self.assertEqual(test_span.t, consume1_span.t)

        # Parent relationships
        self.assertEqual(publish1_span.p, test_span.s)
        self.assertEqual(consume1_span.p, publish1_span.s)
        self.assertEqual(publish2_span.p, consume1_span.s)

        # Span type
        self.assertEqual(publish1_span.k, 2)  # exit
        self.assertEqual(consume1_span.k, 1)  # entry
        self.assertEqual(publish2_span.k, 2)  # exit

        # publish
        self.assertEqual('test.exchange', publish1_span.data["rabbitmq"]["exchange"])
        self.assertEqual('publish', publish1_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(publish1_span.data["rabbitmq"]["address"])
        self.assertEqual('routing.key', publish1_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(publish1_span.stack)
        self.assertTrue(type(publish1_span.stack) is list)
        self.assertGreater(len(publish1_span.stack), 0)

        self.assertEqual('test.exchange', publish2_span.data["rabbitmq"]["exchange"])
        self.assertEqual('publish', publish2_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(publish2_span.data["rabbitmq"]["address"])
        self.assertEqual('another.key', publish2_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(publish2_span.stack)
        self.assertTrue(type(publish2_span.stack) is list)
        self.assertGreater(len(publish2_span.stack), 0)

        # consume
        self.assertEqual('test.exchange', consume1_span.data["rabbitmq"]["exchange"])
        self.assertEqual('consume', consume1_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(consume1_span.data["rabbitmq"]["address"])
        self.assertEqual('routing.key', consume1_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(consume1_span.stack)
        self.assertTrue(type(consume1_span.stack) is list)
        self.assertGreater(len(consume1_span.stack), 0)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(consume1_span.ec)
        self.assertIsNone(publish1_span.ec)
        self.assertIsNone(publish2_span.ec)

    def test_consume_with_ensure_future(self):
        async def run_later(msg):
            # Extract the context from the message (if there is any)
            ctx = async_tracer.extract(opentracing.Format.HTTP_HEADERS, dict(msg.headers))

            # Start a new span to track work that is done processing this message
            with async_tracer.start_active_span("run_later", child_of=ctx) as scope:
                scope.span.set_tag("exchange", msg.exchange_name)
                # print("")
                # print("run_later active scope: %s" % async_tracer.scope_manager.active)
                # print("")
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/")

        def handle_message(msg):
            # print("")
            # print("handle_message active scope: %s" % async_tracer.scope_manager.active)
            # print("")
            async_tracer.inject(async_tracer.active_span.context, opentracing.Format.TEXT_MAP, msg.headers)
            asyncio.ensure_future(run_later(msg))
            msg.ack()

        @asyncio.coroutine
        def test():
            with async_tracer.start_active_span('test'):
                msg1 = asynqp.Message({'consume': 'this'})
                self.exchange.publish(msg1, 'routing.key')

            self.consumer = yield from self.queue.consume(handle_message)
            yield from asyncio.sleep(0.5)
            self.consumer.cancel()

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(6, len(spans))

        publish_span = spans[0]
        test_span = spans[1]
        consume_span = spans[2]
        wsgi_span = spans[3]
        aioclient_span = spans[4]
        run_later_span = spans[5]

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, publish_span.t)
        self.assertEqual(test_span.t, consume_span.t)
        self.assertEqual(test_span.t, aioclient_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(publish_span.p, test_span.s)
        self.assertEqual(consume_span.p, publish_span.s)
        self.assertEqual(aioclient_span.p, run_later_span.s)
        self.assertEqual(run_later_span.p, consume_span.s)
        self.assertEqual(wsgi_span.p, aioclient_span.s)

        # Span type
        self.assertEqual(publish_span.k, 2)  # exit
        self.assertEqual(consume_span.k, 1)  # entry

        # publish
        self.assertEqual('test.exchange', publish_span.data["rabbitmq"]["exchange"])
        self.assertEqual('publish', publish_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(publish_span.data["rabbitmq"]["address"])
        self.assertEqual('routing.key', publish_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(publish_span.stack)
        self.assertTrue(type(publish_span.stack) is list)
        self.assertGreater(len(publish_span.stack), 0)

        # consume
        self.assertEqual('test.exchange', consume_span.data["rabbitmq"]["exchange"])
        self.assertEqual('consume', consume_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(consume_span.data["rabbitmq"]["address"])
        self.assertEqual('routing.key', consume_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(consume_span.stack)
        self.assertTrue(type(consume_span.stack) is list)
        self.assertGreater(len(consume_span.stack), 0)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(consume_span.ec)
        self.assertIsNone(publish_span.ec)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

from __future__ import absolute_import

import copy
import wrapt

from ..log import logger
from .pep0249 import ConnectionFactory

try:
    import psycopg2
    import psycopg2.extras

    cf = ConnectionFactory(connect_func=psycopg2.connect, module_name='postgres')

    setattr(psycopg2, 'connect', cf)
    if hasattr(psycopg2, 'Connect'):
        setattr(psycopg2, 'Connect', cf)

    @wrapt.patch_function_wrapper('psycopg2.extensions', 'register_type')
    def register_type_with_instana(wrapped, instance, args, kwargs):
        args_clone = list(copy.copy(args))

        if (len(args_clone) >= 2) and hasattr(args_clone[1], '__wrapped__'):
            args_clone[1] = args_clone[1].__wrapped__

        return wrapped(*args_clone, **kwargs)

    @wrapt.patch_function_wrapper('psycopg2._json', 'register_json')
    def register_json_with_instana(wrapped, instance, args, kwargs):
        if 'conn_or_curs' in kwargs:
            if hasattr(kwargs['conn_or_curs'], '__wrapped__'):
                kwargs['conn_or_curs'] = kwargs['conn_or_curs'].__wrapped__

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting psycopg2")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2016

from __future__ import print_function
import os
import sys
import logging

logger = None


def get_standard_logger():
    """
    Retrieves and configures a standard logger for the Instana package

    @return: Logger
    """
    standard_logger = logging.getLogger("instana")

    ch = logging.StreamHandler()
    f = logging.Formatter('%(asctime)s: %(process)d %(levelname)s %(name)s: %(message)s')
    ch.setFormatter(f)
    standard_logger.addHandler(ch)
    standard_logger.setLevel(logging.DEBUG)
    return standard_logger


def get_aws_lambda_logger():
    """
    Retrieves the preferred logger for AWS Lambda

    @return: Logger
    """
    aws_lambda_logger = logging.getLogger()
    aws_lambda_logger.setLevel(logging.INFO)
    return aws_lambda_logger

def glogging_available():
    """
    Determines if the gunicorn.glogging package is available

    @return:  Boolean
    """
    package_check = False

    # Is the glogging package available?
    try:
        from gunicorn import glogging
    except ImportError:
        pass
    else:
        package_check = True
    
    return package_check

def running_in_gunicorn():
    """
    Determines if we are running inside of a gunicorn process.

    @return:  Boolean
    """
    process_check = False

    try:
        # Is this a gunicorn process?
        if hasattr(sys, 'argv'):
            for arg in sys.argv:
                if arg.find('gunicorn') >= 0:
                    process_check = True
        elif os.path.isfile("/proc/self/cmdline"):
            with open("/proc/self/cmdline") as cmd:
                contents = cmd.read()

            parts = contents.split('\0')
            parts.pop()
            cmdline = " ".join(parts)

            if cmdline.find('gunicorn') >= 0:
                process_check = True

        return process_check
    except Exception:
        logger.debug("Instana.log.running_in_gunicorn: ", exc_info=True)
        return False


aws_env = os.environ.get("AWS_EXECUTION_ENV", "")
env_is_aws_lambda = "AWS_Lambda_" in aws_env

if running_in_gunicorn() and glogging_available():
    logger = logging.getLogger("gunicorn.error")
elif env_is_aws_lambda is True:
    logger = get_aws_lambda_logger()
else:
    logger = get_standard_logger()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

import os
import sys
import grpc
import time
import tests.apps.grpc_server.stan_pb2 as stan_pb2
import tests.apps.grpc_server.stan_pb2_grpc as stan_pb2_grpc
from concurrent import futures

try:
    from ...helpers import testenv
except ValueError:
    # We must be running from the command line...
    testenv = {}

testenv["grpc_port"] = 10814
testenv["grpc_host"] = "127.0.0.1"
testenv["grpc_server"] = testenv["grpc_host"] + ":" + str(testenv["grpc_port"])


class StanServicer(stan_pb2_grpc.StanServicer):
    """
    gRPC server for Stan Service
    """
    def __init__(self, *args, **kwargs):
        self.server_port = testenv['grpc_port']

    def OneQuestionOneResponse(self, request, context):
        # print("😇:I was asked: %s" % request.question)
        response = """\
Invention, my dear friends, is 93% perspiration, 6% electricity, \
4% evaporation, and 2% butterscotch ripple. – Willy Wonka"""
        result = {'answer': response, 'was_answered': True}
        return stan_pb2.QuestionResponse(**result)

    def ManyQuestionsOneResponse(self, request_iterator, context):
        for request in request_iterator:
            # print("😇:I was asked: %s" % request.question)
            pass

        result = {'answer': 'Ok', 'was_answered': True}
        return stan_pb2.QuestionResponse(**result)

    def OneQuestionManyResponses(self, request, context):
        # print("😇:I was asked: %s" % request.question)
        for count in range(6):
            result = {'answer': 'Ok', 'was_answered': True}
            yield stan_pb2.QuestionResponse(**result)

    def ManyQuestionsManyReponses(self, request_iterator, context):
        for request in request_iterator:
            # print("😇:I was asked: %s" % request.question)
            result = {'answer': 'Ok', 'was_answered': True}
            yield stan_pb2.QuestionResponse(**result)

    def OneQuestionOneErrorResponse(self, request, context):
            # print("😇:I was asked: %s" % request.question)
            raise Exception('Simulated error')
            result = {'answer': "ThisError", 'was_answered': True}
            return stan_pb2.QuestionResponse(**result)

    def start_server(self):
        """
        Function which actually starts the gRPC server, and preps
        it for serving incoming connections
        """
        # declare a server object with desired number
        # of thread pool workers.
        rpc_server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

        # This line can be ignored
        stan_pb2_grpc.add_StanServicer_to_server(StanServicer(), rpc_server)

        # bind the server to the port defined above
        rpc_server.add_insecure_port('[::]:{}'.format(self.server_port))

        # start the server
        rpc_server.start()

        try:
            # need an infinite loop since the above
            # code is non blocking, and if I don't do this
            # the program will exit
            while True:
                time.sleep(60*60*60)
        except KeyboardInterrupt:
            rpc_server.stop(0)
            print('Stan as a Service RPC Server Stopped ...')


if __name__ == "__main__":
    print ("Booting foreground GRPC application...")
    # os.environ["INSTANA_TEST"] = "true"

    if sys.version_info >= (3, 5, 3):
        StanServicer().start_server()
    else:
        print("Python v3.5.3 or higher only")
